import numpy as numpy
import math
from .NonLinearModel import NonLinearModel
from .LorentzModel import LorentzModel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2011 Do Kester, SRON (JAVA code)
#  *    2016 - 2017 Do Kester

class VoigtModel( NonLinearModel ):
    """
    Voigt's Gauss Lorentz convoluted model for line profiles.

    The Voigt function is a convolution of a Gauss and a Lorentz function.
    Physicaly it is the result of thermal and pressure broadening of a spectral
    line.

    The models takes 4 parameters: amplitude, center frequency, half-width of
    the Gaussian, and half-width of the Lorentzian.
    These are initialised to [1, 0, 1, 1].
    Parameters 2 & 3 ( widths ) is always kept positive ( >=0 ).

    Examples
    --------
    >>> voigt = VoigtModel( )
    >>> voigt.setParameters( [5, 4, 1, 2] )
    >>> print( voigt( numpy.arange(  41 , dtype=float ) / 5 ) )      # from [0,8]

    """

    R0 = 146.7
    R1 = 14.67
    RRTPI = 0.56418958              #  = 1/sqrt( pi )
    Y0 = 1.5
    Y0PY0 = Y0 + Y0
    Y0Q = Y0 * Y0
    C = [1.0117281, -0.75197147, 0.012557727, 0.010022008, -0.00024206814, 0.00000050084806]
    S = [1.393237, 0.23115241, -0.15535147, 0.0062183662,  0.000091908299, -0.00000062752596]
    T = [0.31424038, 0.94778839, 1.5976826, 2.2795071, 3.0206370, 3.8897249]
    LN2 = math.log( 2.0 )
    SRLN2 = math.sqrt( LN2 )

    def __init__( self, copy=None, **kwargs ):
        """
        Voigt model.
        <br>
        Number of parameters is 4.

        Parameters
        ----------
        copy : VoigtModel
            to be copied

        """
        param = [1.0, 0.0, 1.0, 1.0]
        names = ["amplitude","center","gausswidth","lorentzwidth"]

        super( VoigtModel, self ).__init__( 4, copy=copy, params=param,
                        names=names, **kwargs )
        if copy is None :
            self.posIndex = [2,3]
            self.nonZero  = [3]

    def copy( self ):
        """ Copy method.  """
        return VoigtModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Note: both width in the parameter array ( items 2 & 3 ) are kept
        strictly positive. I.e. they are changed when upon xdata they are negative.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        if params[2] == 0:
            p = params[[0,1,3]]
            return LorentzModel( ).result( xdata, p )

        rtn = numpy.zeros_like( xdata )
        x = ( xdata - params[1] ) * self.SRLN2 / params[2]
        y = self.SRLN2 * params[3] / params[2]
        for k in range( len( xdata ) ) :
            rtn[k] = self.calculate( x[k], y )

        # Avoid multiple calculations of r0 for the same y
        r0 = self.calculate( 0, y )

        return params[0] * rtn / r0

    def baseName( self ):
        """
        Returns a string representation of the model.
        """
        return str( "Voigt:" )

    def calculate( self, xAxis, y ):
        ## This is legacy code, probably from FORTRAN.
        ## Unfortunately, I dont know where it came from.

        realPart = 0
        RG1 = 1
        RG2 = 1
        RG3 = 1
        YQ = y * y
        YRRTPI = y * self.RRTPI
        ABX = 0.0
        XQ = 0.0
        XLIM0 = self.R0 - y
        XLIM1 = self.R1 - y
        XLIM3 = 3.097 * y - 0.45
        XLIM2 = 6.8 - y
        XLIM4 = 18.1 * y + 1.65
        A0 = 0.0
        D0 = 0.0
        D2 = 0.0
        E0 = 0.0
        E2 = 0.0
        E4 = 0.0
        H0 = 0.0
        H2 = 0.0
        H4 = 0.0
        H6 = 0.0
        P0 = 0.0
        P2 = 0.0
        P4 = 0.0
        P6 = 0.0
        P8 = 0.0
        Z0 = 0.0
        Z2 = 0.0
        Z4 = 0.0
        Z6 = 0.0
        Z8 = 0.0
        D = 0.0
        YF = 0.0
        YPY0 = 0.0
        YPY0Q = 0.0
        XP = numpy.zeros( 6 )
        XM = numpy.zeros( 6 )
        YP = numpy.zeros( 6 )
        YM = numpy.zeros( 6 )
        MQ = numpy.zeros( 6 )
        PQ = numpy.zeros( 6 )
        MF = numpy.zeros( 6 )
        PF = numpy.zeros( 6 )
        ABX = abs( xAxis )
        XQ = ABX * ABX
        if ABX > XLIM0:
            realPart = YRRTPI / ( XQ + YQ )
        elif ABX > XLIM1:
            if RG1 != 0:
                RG1 = 0
                A0 = YQ + 0.5
                D0 = A0 * A0
                D2 = YQ + YQ - 1.0
            D = self.RRTPI / ( D0 + XQ * ( D2 + XQ) )
            realPart = D * y * ( A0 + XQ )
        elif ABX > XLIM2:
            if RG2 != 0:
                RG2 = 0
                H0 = 0.5625 + YQ * ( 4.5 + YQ * ( 10.5 + YQ * ( 6.0 + YQ)) )
                H2 = -4.5 + YQ * ( 9.0 + YQ * ( 6.0 + YQ * 4.0) )
                H4 = 10.5 - YQ * ( 6.0 - YQ * 6.0 )
                H6 = -6.0 + YQ * 4.0
                E0 = 1.875 + YQ * ( 8.25 + YQ * ( 5.5 + YQ) )
                E2 = 5.25 + YQ * ( 1.0 + YQ * 3.0 )
                E4 = 0.75 * H6
            D = self.RRTPI / ( H0 + XQ * ( H2 + XQ * ( H4 + XQ * ( H6 + XQ))) )
            realPart = D * y * ( E0 + XQ * ( E2 + XQ * ( E4 + XQ)) )
        elif ABX < XLIM3:
            if RG3 != 0:
                RG3 = 0
                Z0 = 272.1014 + y * ( 1280.829 + y * ( 2802.870 + y * ( 3764.966 + y * ( 3447.629 + y * ( 2256.981 + y * ( 1074.409 + y * ( 369.1989 + y * ( 88.26741 + y * ( 13.39880 + y)))))))) )
                Z2 = 211.678 + y * ( 902.3066 + y * ( 1758.336 + y * ( 2037.310 + y * ( 1549.675 + y * ( 793.4273 + y * ( 266.2987 + y * ( 53.59518 + y * 5.0)))))) )
                Z4 = 78.86585 + y * ( 308.1852 + y * ( 497.3014 + y * ( 479.2576 + y * ( 269.2916 + y * ( 80.39278 + y * 10.0)))) )
                Z6 = 22.03523 + y * ( 55.02933 + y * ( 92.75679 + y * ( 53.59518 + y * 10.0)) )
                Z8 = 1.496460 + y * ( 13.39880 + y * 5.0 )
                P0 = 153.5168 + y * ( 549.3954 + y * ( 919.4955 + y * ( 946.8970 + y * ( 662.8097 + y * ( 328.2151 + y * ( 115.3772 + y * ( 27.93941 + y * ( 4.264678 + y * 0.3183291))))))) )
                P2 = -34.16955 + y * ( -1.322256 + y * ( 124.5975 + y * ( 189.7730 + y * ( 139.4665 + y * ( 56.81652 + y * ( 12.79458 + y * 1.2733163))))) )
                P4 = 2.584042 + y * ( 10.46332 + y * ( 24.01655 + y * ( 29.81482 + y * ( 12.79568 + y * 1.9099744))) )
                P6 = -0.07272979 + y * ( 0.9377051 + y * ( 4.266322 + y * 1.273316) )
                P8 = 0.0005480304 + y * 0.3183291
            D = 1.7724538 / ( Z0 + XQ * ( Z2 + XQ * ( Z4 + XQ * ( Z6 + XQ * ( Z8 + XQ)))) )
            realPart = D * ( P0 + XQ * ( P2 + XQ * ( P4 + XQ * ( P6 + XQ * P8))) )
        else:
            YPY0 = y + self.Y0
            YPY0Q = YPY0 * YPY0
            realPart = 0.0
            for J in range( 6 ) :
                D = xAxis - self.T[J]
                MQ[J] = D * D
                MF[J] = 1.0 / ( MQ[J] + YPY0Q )
                XM[J] = MF[J] * D
                YM[J] = MF[J] * YPY0
                D = xAxis + self.T[J]
                PQ[J] = D * D
                PF[J] = 1.0 / ( PQ[J] + YPY0Q )
                XP[J] = PF[J] * D
                YP[J] = PF[J] * YPY0

            if ABX <= XLIM4:
                for J in range( 6 ) :
                    realPart = realPart + self.C[J] * ( YM[J] + YP[J] ) - self.S[J] * ( XM[J] - XP[J] )
            else:
                YF = y + self.Y0PY0
                for J in range( 6 ) :
                    realPart = realPart + ( self.C[J] * ( MQ[J] * MF[J] - self.Y0 * YM[J] ) + self.S[J] * YF * XM[J] ) / ( MQ[J] + self.Y0Q ) + ( self.C[J] * ( PQ[J] * PF[J] - self.Y0 * YP[J] ) - self.S[J] * YF * XP[J] ) / ( PQ[J] + self.Y0Q )
                realPart = y * realPart + math.exp( -XQ )
        return realPart

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        if k == 0:
            return self.yUnit
        return self.xUnit





# -*- coding: utf-8 -*-
"""Python/Numpy implementation of Bspline basis functions via Cox - de Boor algorithm."""

from __future__ import division

from functools import partial
import numpy as np

class memoize(object):
    """
       Cache the return value of a method

       This class is meant to be used as a decorator of methods. The return value
       from a given method invocation will be cached on the instance whose method
       was invoked. All arguments passed to a method decorated with memoize must
       be hashable.

       If a memoized method is invoked directly on its class the result will not
       be cached. Instead the method will be invoked like a static method:
       class Obj(object):
           @memoize
           def add_to(self, arg):
               return self + arg
       Obj.add_to(1) # not enough arguments
       Obj.add_to(1, 2) # returns 3, result is not cached

       Script borrowed from here:
       MIT Licensed, attributed to Daniel Miller, Wed, 3 Nov 2010
       http://code.activestate.com/recipes/577452-a-memoize-decorator-for-instance-methods/
    """
    def __init__(self, func):
        self.func = func
    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)
    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(kw.items()))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res



class Bspline():
    """
    Numpy implementation of Cox - de Boor algorithm in 1D

    inputs:
        knot_vector: Python list or Numpy array containing knot vector
                        entries
        order: Order of interpolation, e.g. 0 -> piecewise constant between
                  knots, 1 -> piecewise linear between knots, etc.
    last : bool
        if True the xi equal to the value of the last knot are also taken along

    outputs:
           basis object that is callable to evaluate basis functions at given
           values of knot span
    """

    def __init__(self, knot_vector, order, last=False ):
        """Initialize attributes"""
        self.knot_vector = np.array(knot_vector)
        if last :
            self.last_knot = np.full_like( knot_vector, knot_vector[0] - 1 )
            self.last_knot[-order-1] = knot_vector[-order-1]
        self.last = last
        self.p = order

        #Dummy calls to the functions for memory storage
        self.__call__(0.0)
        self.d(0.0)


    def __basis0(self, xi):
        """Order zero basis"""

        upper =  xi < self.knot_vector[1:]
        if self.last :
            upper = np.any( [upper, xi == self.last_knot[1:]], axis=0 )

        b = np.where(np.all([self.knot_vector[:-1] <= xi, upper], axis=0 ), 1.0, 0.0 )
        return b

    def __basis(self, xi, p, compute_derivatives=False):
        """
        Recursive Cox - de Boor function to compute basis functions and
           optionally their derivatives
        """

        if p == 0:
            return self.__basis0(xi)
        else:
            basis_p_minus_1 = self.__basis(xi, p - 1)

        first_term_numerator = xi - self.knot_vector[:-p]
        first_term_denominator = self.knot_vector[p:] - self.knot_vector[:-p]

        second_term_numerator = self.knot_vector[(p + 1):] - xi
        second_term_denominator = (self.knot_vector[(p + 1):] -
                                   self.knot_vector[1:-p])


        #Change numerator in last recursion if derivatives are desired
        if compute_derivatives and p == self.p:

            first_term_numerator = p
            second_term_numerator = -p

        #Disable divide by zero error because we check for it
        with np.errstate(divide='ignore', invalid='ignore'):
            first_term = np.where(first_term_denominator != 0.0,
                                  (first_term_numerator /
                                   first_term_denominator), 0.0)
            second_term = np.where(second_term_denominator != 0.0,
                                   (second_term_numerator /
                                    second_term_denominator), 0.0)

        return  (first_term[:-1] * basis_p_minus_1[:-1] +
                 second_term * basis_p_minus_1[1:])


    @memoize
    def __call__(self, xi):
        """
        Convenience function to make the object callable.  Also 'memoized'
           for speed.
        """
        return self.__basis(xi, self.p, compute_derivatives=False)

    @memoize
    def d(self, xi):
        """
        Convenience function to compute derivate of basis functions.
           'Memoized' for speed.
        """
        return self.__basis(xi, self.p, compute_derivatives=True)

    def plot(self):
        """
        Convenience function to plot basis functions over full
        range of knots.
        """

        import matplotlib.pyplot as plt

        x_min = np.min(self.knot_vector)
        x_max = np.max(self.knot_vector)

        x = np.linspace(x_min, x_max, num=1000)

        N = np.array([self(i) for i in x]).T;

        for n in N:

            plt.plot(x,n)

        return plt.show()

    def dplot(self):
        """
        Convenience function to plot derivatives of basis functions over
        full range of knots.
        """

        import matplotlib.pyplot as plt

        x_min = np.min(self.knot_vector)
        x_max = np.max(self.knot_vector)

        x = np.linspace(x_min, x_max, num=1000)

        N = np.array([self.d(i) for i in x]).T;

        for n in N:

            plt.plot(x,n)

        return plt.show()


    def __diff_internal(self):
        """
        Differentiate a B-spline once, and return the resulting coefficients and Bspline objects.

        This preserves the Bspline object nature of the data, enabling recursive implementation
        of higher-order differentiation (see `diff`).

        The value of the first derivative of `B` at a point `x` can be obtained as::

            def diff1(B, x):
                terms = B.__diff_internal()
                return sum( ci*Bi(x) for ci,Bi in terms )

        Returns:
            tuple of tuples, where each item is (coefficient, Bspline object).

        See:
            `diff`: differentiation of any order >= 0
        """
        assert self.p > 0, "order of Bspline must be > 0"  # we already handle the other case in diff()

        # https://www.cs.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/B-spline/bspline-derv.html
        #
        t    = self.knot_vector
        p    = self.p
        Bi   = Bspline( t[:-1], p-1 )
        Bip1 = Bspline( t[1:],  p-1 )

        numer1 = +p
        numer2 = -p
        denom1 = t[p:-1]   - t[:-(p+1)]
        denom2 = t[(p+1):] - t[1:-p]

        with np.errstate(divide='ignore', invalid='ignore'):
            ci   = np.where(denom1 != 0., (numer1 / denom1), 0.)
            cip1 = np.where(denom2 != 0., (numer2 / denom2), 0.)

        return ( (ci,Bi), (cip1,Bip1) )


    def diff(self, order=1):
        """
        Differentiate a B-spline `order` number of times.

        Parameters:
        order: int, >= 0

        Returns:
            **lambda** `x`: ... that evaluates the `order`-th derivative of `B` at the point `x`.
        """
        order = int(order)
        if order < 0:
            raise ValueError("order must be >= 0, got %d" % (order))

        if order == 0:
            return self.__call__

        if order > self.p:   # identically zero, but force the same output format as in the general case
            dummy = self.__call__(0.)  # get number of basis functions and output dtype
            nbasis = dummy.shape[0]
            return lambda x: np.zeros( (nbasis,), dtype=dummy.dtype )  # accept but ignore input x

        # At each differentiation, each term maps into two new terms.
        # The number of terms in the result will be 2**order.
        #
        # This will cause an exponential explosion in the number of terms for high derivative orders,
        # but for the first few orders (practical usage; >3 is rarely needed) the approach works.
        #
        terms = [ (1.,self) ]
        for k in range(order):
            tmp = []
            for Ci,Bi in terms:
                tmp.extend( (Ci*cn, Bn) for cn,Bn in Bi.__diff_internal() )  # NOTE: also propagate Ci
            terms = tmp

        # perform final summation at call time
        return lambda x: sum( ci*Bi(x) for ci,Bi in terms )


    def collmat(self, tau, deriv_order=0):
        """
        Compute collocation matrix.

        Parameters:
        tau:
            Python list or rank-1 array, collocation sites
        deriv_order:
            int, >=0, order of derivative for which to compute the collocation matrix.
            The default is 0, which means the function value itself.

        Returns:
        A:
            rank-2 array such that
                A[i,j] = D**deriv_order B_j(tau[i])
            where
                D**k  = kth derivative (0 for function value itself)

        Example:
        If the coefficients of a spline function are given in the vector c, then::

            sum( A[i,:] * c[:] )

        will give a rank-1 array of function values at the sites tau[i] that were supplied
        to `collmat`.

        Similarly for derivatives (if the supplied `deriv_order`> 0).

        """
        # get number of basis functions and output dtype
        dummy = self.__call__(0.)
        nbasis = dummy.shape[0]

        tau = np.atleast_1d(tau)
        if tau.ndim > 1:
            raise ValueError("tau must be a list or a rank-1 array")

        A = np.empty( (tau.shape[0], nbasis), dtype=dummy.dtype )
        f = self.diff(order=deriv_order)
        for i,taui in enumerate(tau):
            A[i,:] = f(taui)

        return np.squeeze(A)



import numpy as numpy
import math

from .ScaledErrorDistribution import ScaledErrorDistribution

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


class GaussErrorDistribution( ScaledErrorDistribution ):
    """
    To calculate a Gauss likelihood.

    For one residual, x, it holds
    .. math::
        f( x ) = 1 / \sqrt( 2 \pi s^2 ) exp( - 0.5 ( x / s )^2 )

    where s is the scale.
    s is a hyperparameter, which might be estimated from the data.

    The scale s is also the sqrt of the variance of this error distribution.

    The function is mostly used to calculate the likelihood L, or easier
    to use log likelihood, logL.
    .. math::
        logL = \log( N / ( \sqrt( 2 \pi ) s )  ) - 0.5 \sum( ( x / s ) ^ 2 )

    Using weights this becomes:
    .. math::
        logL = \log( \sum( w ) / ( \sqrt( 2 \pi ) s )  ) - 0.5 \sum( w ( x / s ) ^ 2 )


    Author       Do Kester.

    """
    LOG2PI = math.log( 2 * math.pi )


    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata, data, weights=None, scale=1.0, limits=None,
                  copy=None ):
        """
        Default Constructor.

        Parameters
        ----------
        xdata : array_like
            input data for the model (independent data)
        data : array_like
            data to be fitted (dependent data)
        weights : array_like
            weights to be used
        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is *not* fixed.

        copy : GaussErrorDistribution
            distribution to be copied.

        """
        super( GaussErrorDistribution, self ).__init__( xdata, data, weights=weights,
                    scale=scale, limits=limits, copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return GaussErrorDistribution( self.xdata, self.data, copy=self )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return True


    #  *********LIKELIHOODS***************************************************
    def logLikelihood( self, model, allpars ) :
        """
        Return the log( likelihood ) for a Gaussian distribution.

        Parameters
        ----------
        model : Model
            to be fitted
        allpars : array_like
            list of all parameters in the problem

        """
        np = model.npchain
        scale = allpars[np]
        res = self.getResiduals( model, allpars[:np] )
        chisq = self.getChisq( res, scale )
        self.ncalls += 1
        return ( - self.sumweight * ( 0.5 * self.LOG2PI + math.log( scale ) ) -
                       0.5 * chisq )

    def getScale( self, model ) :
        """
        Return the noise scale.

        Parameters
        ----------
        mode : Model
            the model involved
        """
        chi = self.getChisq( self.getResiduals( model ), 1.0 )
        return math.sqrt( chi / self.sumweight )

    def getSumRes( self, residual, scale ):
        return self.getChiSq( residual, scale )

    def getChisq( self, residual, scale ):
        """
        Return chisq.

        Sum over the (weighted) squared normalozed residuals

        Parameters
        ----------
        residual : array_like
            the residuals
        scale : float
            hyperparameter of the problem; here it is the noise scale

        """
        res2 = numpy.square( residual )
        if self.weights is not None :
            res2 = res2 * self.weights
        return numpy.sum( res2  ) / ( scale * scale )

    def partialLogL( self, model, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters in fitIndex.

        Parameters
        ----------
        model : Model
            to be fitted
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        self.nparts += 1                    ## counts calls to partialLogL

        np = model.npchain
        param = allpars[:np]
        scale = allpars[np]
        s2 = scale * scale
        res = self.getResiduals( model, param )
        if self.weights is not None :
            resw = res * self.weights
        else :
            resw = res
        dM = model.partial( self.xdata, param )

        dL = numpy.zeros( len( fitIndex ), dtype=float )
        i = 0
        for  k in fitIndex:
            if k < np :
                dL[i] = numpy.sum( resw * dM[:,k] ) / s2
            else :
                dL[i] = ( numpy.sum( res * resw ) / s2 - self.sumweight ) / scale
            i += 1
        return dL

    def hessianLogL( self, model, allpars, fitIndex ) :
        """
        Return the hessian of log( likelihood ) to the parameters in fitIndex.

        ..math::
             hessian = d^2 logL/ dp_i dp_k

        Parameters
        ----------
        model : Model
            to be fitted
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        self.nparts += 1                    ## counts calls to partialLogL

        nh = len( fitIndex )
        np = model.npchain
        param = allpars[:np]
        scale = allpars[np]
        s2 = scale * scale

        fi = fitIndex if fitIndex[-1] != np else fitIndex[-1:]

        nf = len( fi )
        design = model.partial( self.xdata, param )[:,fi]
        design = design.transpose()
        deswgt = design
        if self.weights is None :
            deswgt *= self.weights

        hessian = numpy.zeros( ( nh, nh ), dtype=float )
        hessian[:nf,:nf] = numpy.inner( design, deswgt ) / s2

        if fitIndex[-1] == np :
            hessian[nf,nf] = 2 * ( self.ndata - nf ) / s2

        return hessian

    def __str__( self ) :
        return "Gauss error distribution"



# run with : python3 -m unittest TestKernel2dModel

import unittest
import numpy as numpy
from astropy import units
import math
from numpy.testing import assert_array_almost_equal as assertAAE

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2006 - 2014 Do Kester, SRON (Java code)
#  *  2017        Do Kester


class TestKernel2dModel( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """

    #      * Define x independent variable

    def __init__( self, name ):
        super( TestKernel2dModel, self ).__init__( name )

    def setUp( self ):
        pass

    def tearDown( self ):
        pass

    def testKernel2dModelDefault( self ):
        print( "*******************************************************" )
        print( "*  Kernel2dModel Circular Gauss                       *" )
        print( "*******************************************************" )

        m = Kernel2dModel( )
        self.assertTrue( m.npchain, 4 )

        x1 = numpy.arange( 101, dtype=float ) / 25.0 - 2.02

        x = x1.copy()
        x = numpy.append( x, x1 + 0.01, 0 ).reshape( 101, 2 )
        print( x )

        p = m.parameters
        print( m )
        print( x[40,:] )
        m.testPartial( [x[40,:]], p )

        for k in range( 4 ) :
            print( "%d  %-12s %-12s"%(k, m.getParameterName(k), m.getParameterUnit( k ) ))

        p = numpy.asarray( [1.0,1.01,3.0,1.2], dtype=float )
        part = m.partial( x, p )
        nump = m.numPartial( x, p )
        assertAAE( part, nump, 4 )
        mc = m.copy( )
        mc.parameters = p
        m.parameters = p
        assertAAE( m.result(x ), mc.result( x ) )
        self.assertTrue( isinstance( mc.kernel, Gauss ) )

    def testKernel2dModelCosine( self ):
        print( "*******************************************************" )
        print( "*  Kernel2dModel Elliptic Cosine                      *" )
        print( "*******************************************************" )

        m = Kernel2dModel( kernel=Cosine(), shape=2 )
        self.assertTrue( m.npchain, 5 )

        x1 = numpy.arange( 101, dtype=float ) / 25.0 - 2.02

        x = x1.copy()
        x = numpy.append( x, x1 + 0.01, 0 ).reshape( 101, 2 )

        p = m.parameters
        print( m )
        m.testPartial( [[0.1,0.2]], p )

        for k in range( m.npbase ) :
            print( "%d  %-12s %-12s"%(k, m.getParameterName(k), m.getParameterUnit( k ) ))

        p = numpy.asarray( [1.0,1.01,3.0,1.2,2.0], dtype=float )
        part = m.partial( x, p )
        nump = m.numPartial( x, p )
        assertAAE( part, nump, 4 )
        mc = m.copy( )
        mc.parameters = p
        m.parameters = p
        assertAAE( m.result(x ), mc.result( x ) )
        self.assertTrue( isinstance( mc.kernel, Cosine ) )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestKernel2dModel.__class__ )


if __name__ == '__main__':
    unittest.main()





# -*- coding: utf-8 -*-
"""Utility functions for B-splines, potentially useful for converting MATLAB codes to Python.

CAVEATS:
    - Only a very minimal set of functionality is implemented here.
    - Some technical details differ from the MATLAB equivalents.

Particularly, we use spline order `p` **as-is** instead of MATLAB's `k` parameter (`k` = `p` + 1)
in the function parameters.

Created on Fri Mar 24 13:52:37 2017

@author: Juha Jeronen, juha.jeronen@tut.fi
"""

from __future__ import division

import numpy as np

from . import bspline


def augknt(knots, order):
    """Augment a knot vector.

Parameters:
    knots:
        Python list or rank-1 array, the original knot vector (without endpoint repeats)
    order:
        int, >= 0, order of spline

Returns:
    list_of_knots:
        rank-1 array that has (`order` + 1) copies of ``knots[0]``, then ``knots[1:-1]``,
        and finally (`order` + 1) copies of ``knots[-1]``.

Caveats:
    `order` is the spline order `p`, not `p` + 1, and existing knots are never deleted.
    The knot vector always becomes longer by calling this function.
"""
    if isinstance(knots, np.ndarray)  and  knots.ndim > 1:
        raise ValueError("knots must be a list or a rank-1 array")
    knots = list(knots)  # ensure Python list

    # One copy of knots[0] and knots[-1] will come from "knots" itself,
    # so we only need to prepend/append "order" copies.
    #
    return np.array( [knots[0]] * order  +  knots  +  [knots[-1]] * order )


def aveknt(t, k):
    """Compute the running average of `k` successive elements of `t`. Return the averaged array.

Parameters:
    t:
        Python list or rank-1 array
    k:
        int, >= 2, how many successive elements to average

Returns:
    rank-1 array, averaged data. If k > len(t), returns a zero-length array.

Caveat:
    This is slightly different from MATLAB's aveknt, which returns the running average
    of `k`-1 successive elements of ``t[1:-1]`` (and the empty vector if  ``len(t) - 2 < k - 1``).

"""
    t = np.atleast_1d(t)
    if t.ndim > 1:
        raise ValueError("t must be a list or a rank-1 array")

    n = t.shape[0]
    u = max(0, n - (k-1))  # number of elements in the output array
    out = np.empty( (u,), dtype=t.dtype )

    for j in range(u):
        out[j] = sum( t[j:(j+k)] ) / k

    return out


def aptknt(tau, order):
    """Create an acceptable knot vector.

Minimal emulation of MATLAB's ``aptknt``.

The returned knot vector can be used to generate splines of desired `order`
that are suitable for interpolation to the collocation sites `tau`.

Note that this is only possible when ``len(tau)`` >= `order` + 1.

When this condition does not hold, a valid knot vector is returned,
but using it to generate a spline basis will not have the desired effect
(the spline will return a length-zero array upon evaluation).

Parameters:
    tau:
        Python list or rank-1 array, collocation sites

    order:
        int, >= 0, order of spline

Returns:
    rank-1 array, `k` copies of ``tau[0]``, then ``aveknt(tau[1:-1], k-1)``,
    and finally `k` copies of ``tau[-1]``, where ``k = min(order+1, len(tau))``.
"""
    tau = np.atleast_1d(tau)
    k   = order + 1

    if tau.ndim > 1:
        raise ValueError("tau must be a list or a rank-1 array")

    # emulate MATLAB behavior for the "k" parameter
    #
    # See
    #   https://se.mathworks.com/help/curvefit/aptknt.html
    #
    if len(tau) < k:
        k = len(tau)

    if not (tau == sorted(tau)).all():
        raise ValueError("tau must be nondecreasing")

    # last processed element needs to be:
    #     i + k - 1 = len(tau)- 1
    # =>  i + k = len(tau)
    # =>  i = len(tau) - k
    #
    u = len(tau) - k
    for i in range(u):
        if tau[i+k-1] == tau[i]:
            raise ValueError("k-fold (or higher) repeated sites not allowed, but tau[i+k-1] == tau[i] for i = %d, k = %d" % (i,k))

    # form the output sequence
    #
    prefix = [ tau[0]  ] * k
    suffix = [ tau[-1] ] * k

    # https://se.mathworks.com/help/curvefit/aveknt.html
    # MATLAB's aveknt():
    #  - averages successive k-1 entries, but ours averages k
    #  - seems to ignore the endpoints
    #
    tmp    = aveknt(tau[1:-1], k-1)
    middle = tmp.tolist()
    return np.array( prefix + middle + suffix, dtype=tmp.dtype )


def knt2mlt(t):
    """Count multiplicities of elements in a sorted list or rank-1 array.

Minimal emulation of MATLAB's ``knt2mlt``.

Parameters:
    t:
        Python list or rank-1 array. Must be sorted!

Returns:
    out
        rank-1 array such that
        out[k] = #{ t[i] == t[k] for i < k }

Example:
    If ``t = [1, 1, 2, 3, 3, 3]``, then ``out = [0, 1, 0, 0, 1, 2]``.

Caveat:
    Requires input to be already sorted (this is not checked).
"""
    t = np.atleast_1d(t)
    if t.ndim > 1:
        raise ValueError("t must be a list or a rank-1 array")

    out   = []
    e     = None
    for k in range(t.shape[0]):
        if t[k] != e:
            e     = t[k]
            count = 0
        else:
            count += 1
        out.append(count)

    return np.array( out )


def spcol(knots, order, tau):
    """Return collocation matrix.

Minimal emulation of MATLAB's ``spcol``.

Parameters:
    knots:
        rank-1 array, knot vector (with appropriately repeated endpoints; see `augknt`, `aptknt`)
    order:
        int, >= 0, order of spline
    tau:
        rank-1 array, collocation sites

Returns:
    rank-2 array A such that

        A[i,j] = D**{m(i)} B_j(tau[i])

    where
        m(i) = multiplicity of site tau[i]

        D**k  = kth derivative (0 for function value itself)
"""
    m = knt2mlt(tau)
    B = bspline.Bspline(knots, order)

    dummy = B(0.)
    nbasis = len(dummy)  # perform dummy evaluation to get number of basis functions

    A = np.empty( (tau.shape[0], nbasis), dtype=dummy.dtype )
    for i,item in enumerate(zip(tau,m)):
        taui,mi = item
        f       = B.diff(order=mi)
        A[i,:]  = f(taui)

    return A



import numpy as numpy
from .HyperParameter import HyperParameter
from .Model import Model


from . import Tools

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester


class ErrorDistribution( object ):
    """
    ErrorDistribution defines general methods for a error distribution.

    Error distributions are used to calculate the likelihoods.

    Author       Do Kester.

    Attributes
    ----------
    xdata : array_like
        input data for the model
    data : array_like
        data to be fitted
    weights : array_like
        weights to be used
    hyperpar : HyperParameter
        hyperparameter for the error distribution
    deltaP : float
        delta for calculating numerical derivatives
    ncalls : int
        number of calls to the logLikelihood
    nparts : int
        number of calls to the partial of the logLikelihood
    sumweight : float
        sum over the weights or ndata
    ndata : int
        number of points in data
    hypar : [float]
        list of values for the hyperparameters
    nphypar : int
        number of hyper parameters in this error distribution
    """

    PARNAMES = ["hypar"]

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata, data, weights=None, fixed=None, copy=None ):
        """
        Constructor.

        Parameters
        ----------
        xdata : array_like
            input data for the model
        data : array_like
            data to be fitted
        weights : array_like
            weights to be used
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
        copy : ErrorDistribution
            distribution to be copied.

        """
        super( ErrorDistribution, self ).__init__()
        self.ncalls = 0
        self.nparts = 0
        if copy is None :
            self.xdata = xdata
            self.data = data
            self.weights = weights
            self.hyperpar = []
            self.deltaP = 0.000001
            self.fixed = self.keepFixed( fixed )
        else :                          ### TBC do we need copies here ????
            self.xdata = copy.xdata
            self.data = copy.data
            self.weights = copy.weights
            self.hyperpar = copy.hyperpar
            self.deltaP = copy.deltaP
            self.fixed = self.keepFixed( copy.fixed )

    def copy( self ):
        """ Return copy of this.  """
        return ErrorDistribution( copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        key0 = ["weights", "hyperpar", "fixed"]
        keys = {"xdata": float, "data": float, "weights": float,
                "hyperpar": HyperParameter}
        key1 = {"deltaP": float, "sumweight" : float, "ncalls": int, "nparts": int,
                "fixed": dict }
        if ( Tools.setNoneAttributes( self, name, value, key0 ) or
             Tools.setListOfAttributes( self, name, value, keys ) or
             Tools.setSingleAttributes( self, name, value, key1 ) ) :
            pass
        else :
            raise AttributeError(
                "Object has no attribute " + name + " of type " + str( value.__class__ ) )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == 'sumweight' :            # Return the sum over weight vector.
            if self.hasWeight() :
                self.sumweight = numpy.sum( self.weights )
                return self.sumweight
            else :
                return len( self.data )
        elif name == 'ndata' :
            return len( self.data )         # number of data points/tuples
        elif name == 'nphypar' :
            return len( self.hyperpar )
        elif name == "hypar" :
            return numpy.asarray( [s.hypar for s in self.hyperpar], dtype=float )
        else :
            raise AttributeError( "Unknown attribute " + name )

        return None

    def getScale( self, model ) :
        """
        return default value: 1.0
        """
        return 1.0

    def toSigma( self, scale ) :
        """
        Return default value : scale
        """
        return scale

    def isBound( self ) :
        """
        True when all priors of its (hyper)parameters are bound
        """
        bound = True
        if self.hyperpar is None :
            return bound

        for sp in self.hyperpar :
            bound = bound and sp.prior.isBound()

        return bound

    def acceptWeight( self ):
        """ True if the distribution accepts weights.  """
        return False

    def hasWeight( self ):
        """ Return whether it has weights.  """
        return self.weights is not None


    ### TBD : turn fixed into dictionary ###########################

    def keepFixed( self, fixed=None ) :
        """
        Keeps (hyper)parameters fixed at the provided values.

        1. Repeated calls start from scratch.<br>
        2. Reset with keepFixed( fixed=None )

        Parameters
        ----------
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
        """
        if self.hyperpar is None :
            return None

        # Reset all fixed to False
        for s in self.hyperpar :
            s.fixed = False

        if fixed is None :
            return None

        for f in fixed.keys() :
            self.hyperpar[f].fixed = True
            self.hyperpar[f].hypar = fixed[f]

        return fixed

    def setPriors( self, priors ) :
        """
        Set priors on the hyper parameter(s).

        Parameters
        ----------
        priors  : (list of) Prior
            prior distribution for the hyperparameters
        """
        for spp,pr in zip( self.hyperpar, priors ) :
            spp.setPrior( pr )

    def setLimits( self, limits ) :
        """
        Set limits on the hyper parameter(s).

        Parameters
        ----------
        limits : [low,high]
            low : float or array_like
                low limits
            high : float or array_like
                high limits
        """
        if len( self.hyperpar ) == 1 :
            self.hyperpar[0].setLimits( limits )
            return
        low = limits[0]
        high = limits[1]
        for spp,lo,hi in zip( self.hyperpar, low, high ) :
            spp.setLimits( [lo,hi] )

    def domain2Unit( self, dval, ks ) :
        """
        Return value in [0,1] for the selected parameter.
        Parameters
        ----------
        dval : float
            hyper parameter value in domain
        ks : int
            selecting index
        """
        try :
            return self.hyperpar[ks].domain2Unit( dval )
        except :
            return 0.0

    def unit2Domain( self, uval, ks ) :
        """
        Return domain value for the selected parameter.
        Parameters
        ----------
        uval : float
            unit value of hyper parameter
        ks : int
            selecting index
        """
        try :
            return self.hyperpar[ks].unit2Domain( uval )
        except :
            return 0.0

    #  *********LIKELIHOODS***************************************************

    def logLikelihood( self, model, allpars ):
        """
        Return the log( likelihood ).

        Parameters
        ----------
        model : Model
            to be fitted
        allpars : array_like
            parameters of the problem
        """
        pass

    def partialLogL( self, model, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Parameters
        ----------
        model : Model
            to be fitted
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        return self.numPartialLogL( model, allpars, fitIndex )

    def numPartialLogL( self, model, allpars, fitIndex ) :
        """
        Return d log( likelihood ) / dp, numerically calculated.

        Parameters
        ----------
        model : Model
            to be fitted
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        dL = numpy.zeros( len( fitIndex ), dtype=float )
        p = allpars.copy( )
        i = 0
        for k in fitIndex :
            p[k] = allpars[k] - self.deltaP
            lm = self.logLikelihood( model, p )
            p[k] = allpars[k] + self.deltaP
            lp = self.logLikelihood( model, p )
            dL[i] = 0.5 * ( lp - lm ) / self.deltaP
            i += 1
            p[k]= allpars[k]
        return dL


    def updateLogL( self, model, allpars, parval=None ):
        """"
        Return a update of the log( likelihood ) given a change in a few parameter.

        This method provides the opportunity to optimize the logL calculation.
        Providing this one, automatically provides the previous one.
        For now it just refers to logLikelihood() itself.

        Parameters
        ----------
        model : Model
            to be fitted
        param : array_like
            parameters of the model
        parval : dict of {int : float}
            int index of a parameter
            float (old) value of the parameter
        """
        return self.logLikelihood( model, allpars )

    def setResult( self ):
        pass

    def __str__( self ) :
        return "Error distribution"

    def hyparname( self, k ) :
        """
        Return name of the hyperparameter

        Parameters
        ----------
        k : int
            index of the hyperparameter
        """
        return self.PARNAMES[k]



from __future__ import print_function

import numpy as numpy
from numpy import ndarray
from numpy import float64
from numpy import int64
import math

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017        Do Kester


class Formatter( object ) :
    """
    Format numbers or arrays of numbers in a nice format.

    Attributes
    ----------
    gllen : int
        global line length
    gindent : int
        number of indentation spaces
    gmax : int
        number of items to display
    fmt : dict of {type: str}
        type : "float64" | "int64"
        str  : format string
    """

    def __init__( self, format=None, indent=0, linelength=80, max=5 ):
        """
        Initialize the formatter with new default values.

        Parameters
        ----------
        format : dict {namestring : formatstring }
            name : "float64" or "int64"
            fmt  : " %fmt"
        indent : int
            number of spaces to indent *after* the first line
            Default is 0
        linelength : int
            length of the lines produced
            Default is 80
        max : int or None
            maximum number of items displayed followed by ... if there are more
            None displays all
            Default is 5
        """

#       global gllen, gindent, gmax

        if format is not None :
            self.fmt = { "float64" : " %8.3f", "int64" : " %8d" }
        else :
            for k in format.keys() :
                self.fmt[k] = format[k]

        self.indent = indent
        self.linelen = linelength
        self.max = max


    def formatter( array, format=None, indent=None, linelength=None, max=-1 ) :
        """
        Format a number or an array nicely into a string

        Parameters override defaults given earlier with init().

        Parameters
        ----------
        array : array_like or number
            number or list of numbers or n-dim array of numbers
        format : None or string
            format applying to one item of array
            Default is "8.3f" for float and "8d" for int
        indent : None or int
            number of spaces to indent *after* the first line
            Default is 0
        linelength : None or int
            length of the lines produced
            Default is 80
        max : None or int
            maximum number of items displayed followed by ... if there are more
            None displays all
            Default is 5

        Returns
        -------
        string : containing the formatted array

        """
        if indent is None :
            indent = self.indent
        if linelength is None :
            linelength = self.linelen
        if max is not None and max >= 0 :
            max = self.max
        count = indent
        dent = indent
        llen = linelength
        mx = max
        nwl = False
        sp = 0

        if not isinstance( array, ndarray ) :
            array = numpy.asarray( array )

        if format is None :
            format = self.format[str(array.dtype)]

        fmtlen = len( format % 1 )
        nwl = False

        result = ""
        result = self.recurse( result, array, format, fmtlen, indent, nwl, max )
        return result

    def recurse( self, result, array, format, fmtlen, indent, nwl, max ) :

        if array.size == 1 :
            if count + fmtlen > llen :
                result += ( "\n%s" % self.spaces( sp + indent ) )
                count = indent
            result += ( format % array )
            count += fmtlen
            return result

        shp = array.shape
        if sp > 0 and nwl :
            result += ( "\n%s" % self.spaces( sp + indent ) )
        result += ( "[" )
        nwl = False
        sp += 1
        shp0 = shp[0] if max is None else min( shp[0], max )
        for k in range( shp0 ) :
            result = self.recurse( result, array[k], format, fmtlen, indent, nwl, max )
            nwl = True
        if max is not None and shp[0] > max :
            result += " ..."
        result += ( "]"  )
        count = 0
        sp -=1
        return result

    def spaces( self, int ) :
        return ( " " * int )




import numpy as numpy

from .ConvergenceError import ConvergenceError
from .BaseFitter import BaseFitter
from .IterativeFitter import IterativeFitter
from .IterationPlotter import IterationPlotter
from .GaussErrorDistribution import GaussErrorDistribution
from .LaplaceErrorDistribution import LaplaceErrorDistribution
from .CauchyErrorDistribution import CauchyErrorDistribution
from .PoissonErrorDistribution import PoissonErrorDistribution
from .GenGaussErrorDistribution import GenGaussErrorDistribution


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON ( Java code )
#  *    2016 - 2017 Do Kester

class MaxLikelihoodFitter( IterativeFitter ):
    """
    Base class with methods common to fitters handling ErrorDistributions.

    Author:      Do Kester.

    Attributes
    ----------
    errdis : None | "gauss" | "laplace" | "cauchy" | "poisson" | "gengauss"
        None : Use _ChiSq as function to be minimized
        name : use -logLikelihood as function to be minimized from the named
                errordistribution.
    scale : float
        the (fixed) noise scale
    power : float (2)
        power of errdis (if applicable)

    Raises
    ------
    ConvergenceError    Something went wrong during the convergence if the fit.

    """
    def __init__( self, xdata, model, errdis=None, scale=None, power=2.0,
                    **kwargs ):
        """
        Create a new iterative fitter, providing xdatas and model.

        This is a base class. It collects stuff common to all iterative fitters.
        It does not work by itself.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted

        errdis : None | "gauss" | "laplace" | "cauchy" | "poisson" | "gengauss"
            None : Use _ChiSq as function to be minimized
            name : use -logLikelihood as function to be minimized from the named
                    errordistribution.
        scale : float
            the (fixed) noise scale of errdis (if applicable)
        power : float (2.0)
            the power of errdis ( if applicable)
        kwargs : dict
            Possibly includes keywords from
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        if scale is not None :
            kwargs["fixedScale"] = scale

        super( MaxLikelihoodFitter, self ).__init__( xdata, model, **kwargs )

        self.errdis = errdis
        self.power = power

    def makeFuncs( self, data, weights=None, index=None, ret=3 ) :
        """
        Make connection to the desired func, gradient and hessian.

        Parameters
        ----------
        data : array_like
            the data to be fitted
        weights : array_like or None
            weights on the data
        index : array_like
            indices of the parameters to be fitted.
        ret : 1 or 2 or 3
            return (func), (func,dfunc) or (func,dfunc,hess)
        """
        if not hasattr( self, "errdis" ) or self.errdis is None :
            landscape = _Chisq( self, data, weights, index=index )
            self.isChisq = True
        else :
            scale = 1.0 if self.fixedScale is None else self.fixedScale
            landscape = _LogL( self, data, weights, index=index,
                    errdis=self.errdis, scale=scale, power=self.power )
            self.isChisq = False

#        self.func = landscape.func
#        self.dfunc = landscape.dfunc
#        self.hfunc = landscape.hessian
        self.landscape = landscape

        if ret == 1 :
            return ( landscape.func )

        # Gradient (or fprime or sometimes jacobian ... confusing)
        if not hasattr( self, "userGradient" ) or not self.userGradient :
            dfunc = None                        ## nonexistent, False or None
        elif callable( self.userGradient ) :
            dfunc = landscape.userdfunc         ## callable function
        else :
            dfunc = landscape.dfunc             ## default, True

        if ret == 2 :
            return ( landscape.func, dfunc )

        return ( landscape.func, dfunc, landscape.hessian )


    def getScale( self ) :
        """
        Return the stdev of the noise.
        """
        if isinstance( self.landscape, _LogL ) :
            errdis = self.landscape.errdis
            return errdis.toSigma( errdis.getScale( self.model ) )

        return super( MaxLikelihoodFitter, self ).getScale()


    def getLogLikelihood( self, autoscale=False, var=1.0 ) :
        if self.isChisq :
            return super( MaxLikelihoodFitter, self ).getLogLikelihood(
                        autoscale=autoscale, var=var )

        if autoscale :
            try :
                self.landscape.hypar[0] = self.getScale()
            except IndexError :
                pass
        return - self.landscape.func( self.model.parameters )


    def normalize( self, normdfdp, normdata, weight=1.0 ) :
        """
        Not Implemented.

        Raises
        ------
        NotImplementedError.
        the method is not implemented for MaxLikelihoodFitters

        """
        raise NotImplementedError( "Not implemented for %s." % str( self ) )


    def testGradient( self, par, at, data, weights=None ):
        """
        returns true if the test fails.

        """

        (func,grad) = self.makeFuncs( data, weights=weights, ret=2 )

        df = grad( par )

        dp = 0.001
        pars = par.copy()
        pars[at] -= dp / 2
        r1 = func( pars )
        pars[at] += dp
        r2 = func( pars )
        grat = numpy.subtract( r2, r1 ) / dp

        print( "At %4d  grad : %10.4f  num : %10.4f"%(at, df[at], grat) )

        return ( abs( grat - df[at] ) > 0.001 )

    def __str__( self ):
        return "MaxLikelihoodFitter"


class _Chisq( object ):
    """
    Internal class to provide a chisq landscape.

    It defines
    func (chisq), dfunc (dchisq/dp) and hess (hessian matrix).

    """

    def __init__( self, outer, data, weights, index=None ):
        """
        Parameters
        ----------
        outer : IterativeFitter
            the outer class of this inner class
        data : array_like
            the data
        weights : array_like
            the weights
        index : array_like
            fit index
        """
        self._outer = outer
        self._data = data
        self._weights = weights
        self._index = index

    def func( self, par ):
        """
        Return the function to be minimized.
        In this case
        .. math::
            \chi^2 = \sum( D_i - F(x_i:p) )^2

        """
        param = self._outer.insertParameters( par, index=self._index )
        return self._outer.chiSquared( self._data, params=param,
                    weights=self._weights )

    def userdfunc( self, par ):
        """
        Return an efficient calculation of the gradient.
        """
        return self._outer.userGradient( self._outer.xdata, par, self._data,
                                  weights=self._weights, index=self._index )

    def dfunc( self, par ):
        """
        Return the gradient of func to the parameters p.
        .. math::
            d\chi^2/dp = -2 \sum( D_i - F_i ) dF_i/dp
        """
        param = self._outer.insertParameters( par, index=self._index )
        res = numpy.subtract( self._outer.model.result( self._outer.xdata, param ), self._data )
        if self._weights is not None:
            res = numpy.multiply( res, self._weights )
        desmat = self._outer.getDesign( params=param, index=self._index )

        return 2 * numpy.inner( desmat.transpose(), res )

    def hessian( self, par ):
        """
        Return the Hessian matrix.
        """
        return self._outer.getHessian( params=par, weights=self._weights, index=self._index )

class _LogL( _Chisq ) :
    """
    Internal class to provide a log Likelihood landscape.

    It defines
    func (-logL), dfunc (dlogL/dp) and hess (hessian matrix).

    """
    def __init__( self, outer, data, weights, index=None, errdis='gauss',
                    scale=1.0, power=2.0 ):
        """
        Parameters
        ----------
        outer : IterativeFitter
            the outer class of this inner class
        data : array_like
            the data
        weights : array_like
            the weights
        index : array_like
            fit index
        errdis : "gauss" | "laplace" | "cauchy" | "poisson" | "gengauss"
            errordistribution to be used to calculate logLikelihood
        scale : float (1.0)
            scale if applicable in errdis
        power : float (2.0)
            power if applicable in errdis
        """

        super( _LogL, self ).__init__( outer, data, weights, index=None )

        if errdis == 'gauss' :
            self.errdis = GaussErrorDistribution( outer.xdata, data,
                    weights=weights, scale=scale )
            self.hypar = [scale]
        elif errdis == 'laplace' :
            self.errdis = LaplaceErrorDistribution( outer.xdata, data,
                    weights=weights, scale=scale )
            self.hypar = [scale]
        elif errdis == 'cauchy' :
            self.errdis = CauchyErrorDistribution( outer.xdata, data, scale=scale )
            self.hypar = [scale]
        elif errdis == 'poisson' :
            self.errdis = PoissonErrorDistribution( outer.xdata, data )
            self.hypar = []
        elif errdis == 'gengauss' :
            self.errdis = GenGaussErrorDistribution( outer.xdata, data,
                    weights=weights, scale=scale, power=power )
            self.hypar = [scale, power]
        else :
            raise ValueError( "Unknown errordistribution %s."%errdis )

    def func( self, par ) :
        param = self._outer.insertParameters( par, index=self._index )
        param = numpy.append( param, self.hypar )
        return - self.errdis.logLikelihood( self._outer.model, param )


    def dfunc( self, par ) :
        param = self._outer.insertParameters( par, index=self._index )
        param = numpy.append( param, self.hypar )

        return - self.errdis.partialLogL( self._outer.model, param, self._index )

    def hessian( self, par ) :
        param = self._outer.insertParameters( par, index=self._index )
        param = numpy.append( param, self.hypar )

        return - self.errdis.hessianLogL( self._outer.model, param, self._index )





import math

from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class UniformPrior( Prior ):
    """
    Uniform prior distribution, for location parameters.

    A uniform prior is a improper prior ( i.e. its integral is unbound ).
    Because of that it always needs limits, low and high, such that
    -Inf < low < high < +Inf.
    .. math::
        Pr( x ) = 1 / ( high - low )   if low < x < high else 0

    domain2Unit: u = ( d - lo ) / range
    unit2Domain: d = u * range + lo

    Hidden Attributes
    -----------------
    _range : float
        valid range ( highLimit - lowLimit )

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, limits=None, prior=None ):
        """ Constructor.  """
        super( UniformPrior, self ).__init__( limits=limits, prior=prior )

    def copy( self ):
        """ Return a (deep) copy of itself. """
        return UniformPrior( prior=self, limits=[self.lowLimit,self.highLimit] )

    def __setattr__( self, name, value ) :
        if name == "_range" :
            object.__setattr__( self, name, value )
        else :
            super( UniformPrior, self ).__setattr__( name, value )

        if name == "lowLimit" or name == "highLimit" :
            self._range = self.highLimit - self.lowLimit

    def getIntegral( self ) :
        """
        Return integral of UniformPrior from lowLimit to highLimit.
        """
        return self._range

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Uniform distribution.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        if math.isinf( self._range ) :
            raise AttributeError( "Limits are needed for UniformPrior" )
        return 0 if self.isOutOfLimits( dval ) else ( dval - self.lowLimit ) / self._range

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Uniform distribution.

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        if math.isinf( self._range ) :
            raise AttributeError( "Limits are needed for UniformPrior" )
        return uval * self._range + self.lowLimit

    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        if math.isinf( self._range ) :
            raise AttributeError( "Limits are needed for UniformPrior" )

        return 0.0 if self.isOutOfLimits( x ) else 1.0 / self._range


# logResult has no better definition than the default: just take the math.log of result.
# No specialized method here.

    def partialLog( self, p ):
        """
        Return partial derivative of log( Prior ) wrt parameter.

        Parameters
        ----------
        p : float
            the value

        """
        return math.nan if self.isOutOfLimits( p ) else 0

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return self.hasLowLimit( ) and self.hasHighLimit( )

    def __str__( self ):
        """ Return a string representation of the prior.  """
        return str( "UniformPrior " + ( "unbound." if not self.isBound( )
                else ( "between %.2f and %.2f"%( self.lowLimit, self.highLimit ) ) ) )

#      * End of UniformPrior






import unittest
import numpy as np
import math
import matplotlib.mlab as mlab
import matplotlib.pyplot as plt
from BayesicFitting import ExponentialPrior
from BayesicFitting import LaplacePrior
from BayesicFitting import JeffreysPrior
from BayesicFitting import UniformPrior
from BayesicFitting import GaussPrior
from BayesicFitting import CauchyPrior

from BayesicFitting import Prior
from BayesicFitting import Formatter
from Formatter import formatter as fmt

class PlotTestPrior( unittest.TestCase  ) :

    def testPlotUniform( self ) :
        # example data
        u = np.random.rand( 10000 )
        pr = UniformPrior( limits=[-10,10] )
        d = [pr.unit2Domain( v ) for v in u]
        x = np.asarray( [-12,-10,-10,10,10,12], dtype=float )
        a = 1.0 / pr._range
        y = np.asarray( [0,0,a,a,0,0], dtype=float )
        self.histo( d, pr, fun=(x,y) )

    def testPlotJeffreys( self ) :
        # example data
        u = np.random.rand( 10000 )
        pr = JeffreysPrior( limits=[0.2,9] )
        d = [pr.unit2Domain( v ) for v in u]
        x = np.arange( 51 ) / 5
        y = np.zeros( 51, dtype=float )
        y[1:46] = 1.0 / ( ( math.log( pr.highLimit ) - math.log( pr.lowLimit ) ) * x[1:46] )
        self.histo( d, pr, fun=(x,y) )

    def testPlotExponential( self ) :
        # example data
        u = np.random.rand( 10000 )
        pr = ExponentialPrior( scale=3 )
        d = [pr.unit2Domain( v ) for v in u]
        x = np.arange( 51 ) * pr.scale / 5
        y = np.exp( - x / pr.scale ) / pr.scale
        self.histo( d, pr, fun=(x,y) )

    def testPlotLaplace( self ) :
        # example data
        u = np.random.rand( 10000 )
        pr = LaplacePrior( scale=4 )
        d = [pr.unit2Domain( v ) for v in u]
        x = ( np.arange( 101 ) - 50 ) * pr.scale / 5
        y = np.exp( - np.abs( x ) / pr.scale ) / ( 2 * pr.scale )
        self.histo( d, pr, fun=(x,y) )

    def testPlotGauss( self ) :
        # example data
        u = np.random.rand( 10000 )
        pr = GaussPrior( scale=4 )
        d = [pr.unit2Domain( v ) for v in u]
        x = ( np.arange( 101 ) - 50 ) * pr.scale / 5
        y = [pr.result( v )/4 for v in x]
        self.histo( d, pr, fun=(x,y) )

    """
    def testPlotGaussLut( self ) :
        # example data
        u = np.random.rand( 10000 )
        pr = GaussLutPrior( scale=4, nl=128 )
#        print( fmt( pr.u2d, max=None ) )
#        print( fmt( pr.d2u, max=None ) )
        d = [pr.unit2Domain( v ) for v in u]
        x = ( np.arange( 101 ) - 50 ) * pr.scale / 5
        y = [pr.result( v )/4 for v in x]
        self.histo( d, pr, fun=(x,y) )
    """
    def testPlotCauchy( self ) :
        # example data
        u = np.random.rand( 10000 )
        pr = CauchyPrior( scale=4 )
        d = np.asarray( [pr.unit2Domain( v ) for v in u] )
        # There are too many far outliers in a Cauchy
        q = np.where( d < -50 )
        d[q] = -50
        q = np.where( d > 50 )
        d[q] = 50
        x = ( np.arange( 101 ) - 50 ) * pr.scale / 5
        y = [pr.result( v ) for v in x]
        self.histo( d, pr, fun=(x,y) )


    def histo( self, x, pr, fun=None ) :
        print( pr )
        print( fmt( x ) )
        print( fmt( fun[0] ) )
        print( fmt( fun[0] ) )

        num_bins = 50
        # the histogram of the data
        n, bins, patches = plt.hist(x, num_bins, normed=1, facecolor='green', alpha=0.5)
        # add a 'best fit' line
        if fun is not None :
            plt.plot( fun[0], fun[1], 'r--')
        plt.xlabel('Error')
        plt.ylabel('Probability')
        plt.title('Histogram of ' + str( pr ) )

        # Tweak spacing to prevent clipping of ylabel
        plt.subplots_adjust(left=0.15)
        plt.show()

if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
import math
from . import Tools
from .Formatter import formatter as fmt

from .Model import Model
#from Problem import Problem

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2008 - 2014 Do Kester, SRON (Java code)
#  *    2017        Do Kester

class Sample( object ):
    """
    Sample is weighted random draw from a Posterior distribution as
    provided by a Sampler

    Each Sample maintains 5 attributes

    Attributes
    ----------
    id : int
        identification number
    parent : int
        id of the parent (-1 for Adam/Eve)
    model : Model
        the model being used
    logL : float
        log Likelihood = log Prob( data | params )
    logW : float
        log Weights of the likelihood.
            SUM( log( W * L ) = logZ (TBC)
            log( SUM( W * L ) = logZ (evidence)
    allpars : array_like
        list of parameters and hyperparameters
    fitIndex : array_like or None
        list of (super)parameters to be fitted.
    parameters : array_like (read only)
        parameters (of the model)
    hyperpars : array_like (read only)
        list of hyper parameters (of the error distribution)


    Author       Do Kester

    """

    def __init__( self, id, parent, model, errdis=None, fitIndex=None, copy=None ):
        """
        Constructor.

        Either errdis or copy is obligatory.

        Parameters
        ----------
        id : int
            id of the sample
        parent : int
            id of the parent (-1 for Adam/Eve)
        model : Model
            the model being used. Parameters are copied from this model.
        errdis : ErrorDistribution
            to get info about super parameters
        fitIndex : array_like
            list of indices in allpars that need fitting
        copy : Sample
            the sample to be copied

        """
        self.id = id
        self.parent = parent

        if copy is None :
            self.model = model
            allpars = model.parameters
            npars = len( model.parameters )
            if errdis.nphypar > 0 :
                allpars = numpy.append( allpars, errdis.hypar )
            self.allpars = allpars

            if fitIndex is not None :
                self.fitIndex = fitIndex                # no copy.
            else :
                self.fitIndex = numpy.arange( npars )
                nh = -1
                for s in errdis.hyperpar :
                    if not s.isFixed and s.isBound() :
                        self.fitIndex = numpy.append( self.fitIndex, [nh] )
                    nh -= 1
            self.logL = 0.0
            self.logW = 0.0
        else :
            self.model = copy.model.copy()
            self.allpars = copy.allpars.copy()
            self.fitIndex = copy.fitIndex.copy()
            self.logL = copy.logL
            self.logW = copy.logW

    def copy( self ):
        """
        Copy.

        The copy points to the same instance of model.
        """
        return Sample( self.id, self.parent, self.model, copy=self )

    def __getattr__( self, name ) :
        """
        Return the value of one of `parameters`, `scale`,

        """
        if name == "weight" :
            return math.exp( self.logW )
        elif name == "parameters" :
            return self.allpars[:self.model.npchain]
        elif name == "hypars" :
            if len( self.allpars ) > self.model.npchain :
                return self.allpars[self.model.npchain:]
            else :
                return None
        else :
            raise AttributeError( "Unknown attribute " + name )

        return None

    def __setattr__( self, name, value ) :
        """
        Set attributes.
        """
        if name == "allpars" :
            object.__setattr__( self, name, value )
            return

#        key1 = {"id" : int, "parent" : int, "model": (Model,Problem),
        key1 = {"id" : int, "parent" : int, "model": Model,
                "logL" : float, "logW" : float }
        key2 = {"fitIndex" : int }
        if ( Tools.setSingleAttributes( self, name, value, key1 ) or
             Tools.setListOfAttributes( self, name, value, key2 ) ) :
            pass
        else :
            raise AttributeError(
                "Object has no attribute " + name + " of type " + str( value.__class__ ) )

    def __str__( self ) :
        return str( "Sample: %3d parent: %3d model: %s logL: %10.2f logW: %10.2f"%
                    ( self.id, self.parent, self.model.shortName(), self.logL, self.logW ) )


    def check( self, nhyp=0 ) :
#        print( len( self.allpars ), self.model.npchain, nhyp )
        if not len( self.allpars ) == ( self.model.npchain + nhyp ) :
            raise ValueError( "Sample inconsistent parameter length : %d is not ( %d + %d )" %
                ( len( self.allpars ), self.model.npchain, nhyp ) )
        if nhyp > 0 and self.allpars[-nhyp] <= 0 :
            raise ValueError( "Sample has non-positive hyperparameter: %f" % self.allpars[-nhyp] )



# run with : python3 -m unittest TestLMFitter

import unittest
import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import math

import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import LevenbergMarquardtFitter as LMFitter

from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2004 - 2014 Do Kester, SRON (Java code)
#  *  2017        Do Kester

class TestLMFitter( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def makeData( self, bg=False ) :

        p = [3.2, -0.1, 0.3, 1.1, 2.1]
        ss = 0.3

        ndata = 101
        x = numpy.linspace( -1, +1, ndata, dtype=float )
        y = ( x - p[1] ) / p[2]
        numpy.random.seed( 3456 )
        y = p[0] * numpy.exp( -y * y ) + ss * numpy.random.randn( ndata )
        if not bg :
            return (x,y,p[:3])

        y += ( p[3] + p[4] * x )
        return (x,y,p)

    def eq( self, a, b, eps=1e-10 ) :
        if ( a + b ) != 0 :
            return abs( a - b ) / abs( a + b ) < eps
        else :
            return abs( a - b ) < eps


################################################################################
    def plot1( self ):
        self.test1( plot=True )

    def test1( self, plot=False ):

        x,y,p1 = self.makeData()

        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing Nonlinear Fitters: LevenbergMarquardt (lm)" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        modl1 = GaussModel( )

        amfit = LMFitter( x, modl1 )

        print( fmt( p1 ) )
        par1 = amfit.fit( y )

        print( fmt( par1 ) )
        assertAAE( par1, p1, 1 )

        x,z,p2 = self.makeData( bg=True )

        modl2 = GaussModel( )
        modl2.addModel( PolynomialModel(1) )

        modl2.parameters = numpy.append( par1, [0,0] )
        lmfit = LMFitter( x, modl2 )

        par2 = lmfit.fit( z )
        print( fmt( p2 ) )
        print( fmt( par2 ) )

        print( "chisq1 = ", amfit.chisq, "  chisq2 = ", lmfit.chisq )

        assertAAE( par2, p2, 1 )

        if plot :
            xx = numpy.linspace( -1, +1, 1001 )
            plt.plot( x, y, 'k+' )
            plt.plot( xx, modl1.result( xx ), 'k-' )
            plt.plot( x, z, 'r+' )
            plt.plot( xx, modl2.result( xx ), 'r-' )


        print( fmt( amfit.hessian ) )
        print( fmt( amfit.design[-4:,:], max=None ) )


        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing LevenbergMarquardt (normalized)" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        modl1 = GaussModel( )

        bmfit = LMFitter( x, modl1 )

        conpr1 = numpy.asarray( [1.0,0.0,0.0] )

        bmfit.normalize( conpr1, p1[0], weight=10.0 )

        print( "Normweight = ", bmfit.normweight )

        print( fmt( p1 ) )

        par3 = bmfit.fit( y )
        print( fmt( par1 ) )
        print( fmt( par3 ) )
        print( fmt( amfit.chisq ), fmt( bmfit.chisq ) )

        hes1 = amfit.getHessian( params=par3 )
        hes3 = bmfit.getHessian( params=par3 )
        print( fmt( hes1 ) )
        print( fmt( hes3 ) )

        f = bmfit.normweight
        for h1,h3 in zip( hes1.flat, hes3.flat ) :
            self.assertTrue( abs( h1 + f - h3 ) < 1e-8 )
            f = 0.0

        if plot :
            plt.plot( xx, modl1.result( xx ), 'g-' )
            plt.show()


    def test4( self, plot=False ) :

        self.normalizetest( Fitter )
        self.normalizetest( QRFitter )
        self.normalizetest( CurveFitter )
        self.normalizetest( LMFitter )
        self.assertRaises( NotImplementedError, self.normalizetest, AmoebaFitter )
#        self.assertRaises( NotImplementedError, self.normalizetest, PowellFitter )

    def normalizetest( self, fitter ) :

        p = numpy.asarray( [2.0, 1.3] )
        x = numpy.asarray( [1.0, 1.3, 1.5, 1.8, 2.0] )
        numpy.random.seed( 12345 )
        y = p[0] + p[1] * x + 0.5 * numpy.random.randn( 5 )

        m = PolynomialModel( 1 )

        ftr = fitter( x, m )
        print( "=============================================================" )
        print( str( ftr ) )
        print( fmt( x ) )
        print( fmt( y ) )

        par = ftr.fit( y )
        print( fmt( p ) )
        print( fmt( par ) )

        conpr = numpy.asarray( [1.0,0.0] )
        for w in [0,1,10,100,1000] :
            m2 = m.copy()
            ftr = fitter( x, m2 )

            ftr.normalize( conpr, p[0], weight=w )
            par = ftr.fit( y )
            print( fmt( w ), fmt( par ), fmt( ftr.chisq ) )

        self.assertTrue( abs( p[0] - par[0] ) < 1e-3 )

        print( fmt( ftr.hessian ) )
#        print( fmt( ftr.getVector( y ) ) )




    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestLMFitter.__class__ )





# run with : python3 -m unittest TestCombiModel

import unittest
import numpy as numpy
from numpy.testing import assert_array_equal as assertAE
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import math

from BayesicFitting import *
from StdTests import stdModeltest

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestCombiModel( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """

    def plotall( self ) :
        self.test1()
        self.test2( plot=True )
        self.test3( plot=True )
        self.test4( plot=True )
        self.test5( plot=True )
        self.test6( plot=True )

    def test1( self ):
        print( "\n   CombiModel Test 1\n" )
        gm = GaussModel( )
        self.assertRaises( ValueError, CombiModel, gm, 3, oper="xx" )
        cm = CombiModel( gm, 3 )
        self.assertRaises( ValueError, cm.setCombi, [1,2,3] )
        self.assertRaises( ValueError, cm.setCombi, {3:[1,2,3]} )
        self.assertRaises( ValueError, cm.setCombi, {1:[1,2]} )

    def test2( self, plot=False ):
        print( "\n   CombiModel Test 2\n" )
        gm = GaussModel( )
        pars = numpy.linspace( 0.0, 1.0, 3, dtype=float )
        gm.parameters = pars

        cm = CombiModel( gm, 3 )

        self.assertTrue( cm._deep == 1 )
        print( str( cm._head ) )
        self.assertTrue( str( cm._head ) == "Combi of 3 times Gauss" )
        self.assertTrue( cm._next is None )
        self.assertTrue( cm._npchain == 9 and cm.npchain == 9 )
        assertAE( cm.addindex, [] )
        assertAE( cm.addvalue, [] )
        assertAE( cm.mulindex, [] )
        assertAE( cm.mulvalue, [] )
        assertAE( cm.expandindex, numpy.arange( 9, dtype=int ) )
        assertAE( cm.select, numpy.arange( 9, dtype=int ) )
        assertAE( cm.parameters, [0.0,0.5,1.0]*3 )
        p = numpy.ones( 9, dtype=float )
        p[1:9:3] /= 2
        p[2:9:3] /= 10
        stdModeltest( cm, p, plot=plot )

    def test3( self, plot=False ):
        print( "\n   CombiModel Test 3\n" )
        gm = GaussModel( )
        pars = numpy.linspace( 0.0, 1.0, 3, dtype=float )
        gm.parameters = pars
        cm = CombiModel( gm, 4, addCombi={1:[0,0.4,0.8,1.3], 2:[0,0,0,0]} )

        cm.xUnit = units.m
        cm.yUnit = units.kg

        p = numpy.array( [1.0, -0.6, 0.1, 1.0, 1.0, 1.0] )

        stdModeltest( cm, p, plot=plot )

    def test4( self, plot=False ):
        print( "\n   CombiModel Test 4\n" )
        gm = GaussModel( )
        pars = numpy.linspace( 0.0, 1.0, 3, dtype=float )
        gm.parameters = pars
        cm = CombiModel( gm, 4, addCombi={1:[0,0.4,0.8,1.3]}, mulCombi={2:[1,1,1,1]} )

        cm.xUnit = units.m
        cm.yUnit = units.kg

        p = numpy.array( [1.0, -0.6, 0.1, 1.0, 1.0, 1.0] )

        stdModeltest( cm, p, plot=plot )

    def test5( self, plot=False ):
        print( "\n   CombiModel Test 5\n" )
        gm = EtalonModel( )
        pars = numpy.asarray( [1.0,0.5, 10.0, 0.0], dtype=float )
        gm.parameters = pars
        cm = CombiModel( gm, 2, oper="mul", addCombi={2:[0,1]} )

        cm.xUnit = units.m
        cm.yUnit = units.kg

        p = numpy.asarray( [1.0,0.5, 10.0, 0.0, 1.0, 0.4, 0.0] )
        stdModeltest( cm, p, plot=plot )

    def test6( self, plot=False ):
        print( "\n   CombiModel Test 6\n" )
        gm = EtalonModel( )
        pars = numpy.asarray( [1.0,0.5, 10.0, 0.0], dtype=float )
        gm.parameters = pars
        cm = CombiModel( gm, 2, mulCombi={1:[1,1.4]}, addCombi={2:[0,3]} )

        cm.xUnit = units.m
        cm.yUnit = units.kg

        p = numpy.asarray( [1.0,0.5, 10.0, 0.0, 1.0, 0.4, 0.0] )

        stdModeltest( cm, p, plot=plot )



    def suite( cls ):
        return unittest.TestCase.suite( TestCombiModel.__class__ )






import numpy as numpy
from astropy import units
import math
from . import Tools

from .NonLinearModel import NonLinearModel
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class PadeModel( NonLinearModel ):
    """
    General Pade model of arbitrary degrees in numerator and denominator.
    .. math::
        f( x:p ) = \sum p_n * x^n / ( \sum p_{num+1+k} * x^k )

    where the sum in the numerator is over n running from 0 to num ( inclusive )
    and the sum in the denominator is over k running from 0 to den ( inclusive )

    At least one parameter needs to be fixed, otherwise the model is
    degenerate in its parameters.
    By default the first parameter of the denominator (p_{num+1})
    is fixed to 1.0.

    All parameters are initialized at 0. It is a non-linear model.

    Beware of the poles where the denominator equals zero.

    Author:      Do Kester

    Examples
    --------
    >>> pade = PadeModel( 3, 1 )            # 3rd degree polynomial
    >>> print pade.getNumberOfParameters( )        # 5
    5

    Category:    mathematics/Fitting

    Attributes
    ----------
    num : int
        order of the polynomial in the numerator
    den : int
        order of the polynomial in the denominator

    """
    PARNAMES = ["numer_", "denom_"]

    def __init__( self, num, den, copy=None, fixed=None, **kwargs ):
        """
        Pade of a certain degree in numerator and denominator.

        The number of parameters is ( num + den + 1 )

        Parameters
        ----------
        num : int
            the degree of the polynomial in the numerator.
        den : int
            the degree of the polynomial in the denominator.
        copy : PadeModel
            model to be copied
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently.
                    Default {num+1 : 1.0}
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.

        """
        names  = ["%s%d" % (self.PARNAMES[0], k) for k in range( num + 1 )]
        names += ["%s%d" % (self.PARNAMES[1], k) for k in range( den + 1 )]

        if fixed is None :
            fixed = { num + 1: 1.0 }
        np = num + den + 2
        params = numpy.zeros( np, dtype=float )
        params[num+1] = 1.0
        super( PadeModel, self ).__init__( np, params=params, copy=copy,
                        names=names, fixed=fixed, **kwargs )

        self.num = num
        self.den = den

    def copy( self ):
        """ Copy method.  """
        return PadeModel( self.num, self.den, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: num, den

        """
        dind = {'num': int, 'den': int }
        if not Tools.setSingleAttributes( self, name, value, dind ):
            super( PadeModel, self ).__setattr__( name, value )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model.

        """
        nm = numpy.zeros_like( xdata )
        xx = numpy.ones_like( xdata )
        k = 0
        while k <= self.num :
            nm += params[k] * xx
            xx *= xdata
            k += 1

        dn = numpy.zeros_like( xdata )
        xx = numpy.ones_like( xdata )
        while k < self.npmax :
            dn += params[k] * xx
            xx *= xdata
            k += 1

        return nm / dn

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the xdata values.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        nxdata = Tools.length( xdata )
        partial = numpy.zeros( ( nxdata, self.npmax ), dtype=float )

        xx = numpy.ones( nxdata )
        nm = numpy.zeros_like( xdata )
        k = 0
        while k <= self.num:
            partial[:,k] = xx
            nm += params[k] * xx
            xx *= xdata
            k += 1

        xx = numpy.ones( nxdata )
        dn = numpy.zeros_like( xdata )
        while k < self.npmax :
            partial[:,k] = -xx * nm
            dn += params[k] * xx
            xx *= xdata
            k += 1

        k = 0
        while k <= self.num :
            partial[:,k] /= dn
            k += 1
        while k < self.npmax :
            partial[:,k] /= dn * dn
            k += 1

        if parlist is None :
            return partial

        return partial[:,parlist]

    def baseDerivative( self, xdata, params ):
        """
        Returns the partials at the xdata values.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model.

        """
        nm = numpy.zeros_like( xdata )
        dn = numpy.zeros_like( xdata )
        dnm = numpy.zeros_like( xdata )
        ddn = numpy.zeros_like( xdata )

        xx = numpy.ones_like( xdata )
        nm += params[0] * xx
        k = 1
        while k <= self.num:
            dnm += k * params[k] * xx
            xx *= xdata
            nm += params[k] * xx
            k += 1

        xx = numpy.ones_like( xdata )
        dn += params[k] * xx
        k += 1
        while k < self.npmax :
            ddn += ( k - self.num - 1 ) * params[k] * xx
            xx *= xdata
            dn += params[k] * xx
            k += 1

        return ( dnm * dn - ddn * nm ) / ( dn * dn )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        bn = "Pade: f( x:p ) = ( p_0"
        if self.num > 0 :
            bn += " + p_1 * x"
        for k in range( 2, self.num + 1 ) :
            bn += " + p_%d * x^%d"%(k,k)
        bn += " ) / ( p_%d" % (self.num+1)
        if self.den > 0 :
            bn += " + p_%d * x"%(self.num+2)
        for k in range( 3, self.den + 2 ) :
            bn += " + p_%d * x^%d"%(self.num+k, k-1)
        bn += " )"
        return bn

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Parameters
        ----------
        k : int
            parameter number.

        """
        if k <= self.num:
            return self.yUnit / ( self.xUnit ** k )
        k -= ( self.num + 1 )
        return self.xUnit ** (-k)





import numpy as numpy
import warnings

# unconstrained minimization
from scipy.optimize import minimize

from .MaxLikelihoodFitter import MaxLikelihoodFitter
#from ConvergenceError import ConvergenceError

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2007 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester
#  *

class ScipyFitter( MaxLikelihoodFitter ):
    """
    Unified interface to the Scipy minimization module `minimize`, to fit data to a model.


    For documentation see scipy.org->Docs->Reference Guide->optimization and root finding.
    scipy.optimize.minimize


    Examples
    --------
    # assume x and y are Double1d data arrays.
    >>> x = numpy.arange( 100, dtype=float ) / 10
    >>> y = numpy.arange( 100, dtype=float ) / 122          # make slope
    >>> y += 0.3 * numpy.random.randn( 100 )                # add noise
    >>> y[9:12] += numpy.asarray( [5,10,7], dtype=float )   # make some peak
    >>> gauss = GaussModel( )                               # Gaussian
    >>> gauss += PolynomialModel( 1 )                       # add linear background
    >>> print( gauss.npchain )
    >>> cgfit = ConjugateGradientFitter( x, gauss )
    >>> param = cgfit.fit( y )
    >>> print( len( param ) )
    5
    >>> stdev = cgfit.stdevs
    >>> chisq = cgfit.chisq
    >>> scale = cgfit.scale                                 # noise scale
    >>> yfit  = cgfit.getResult( )                          # fitted values
    >>> yband = cgfit.monteCarloEoor( )                         # 1 sigma confidence region

    Notes
    -----
    1. CGF is <b>not</b> guaranteed to find the global minimum.
    2. CGF does <b>not</b> work with fixed parameters or limits

    Attributes
    ----------
    gradient : callable gradient( par )
        User provided method to calculate the gradient of chisq.
        It can be used to speed up the dotproduct calculation, maybe
        because of the sparseness of the partial.
        default  dotproduct of the partial with the residuals
    tol : float (1.0e-5)
        Stop when the norm of the gradient is less than tol.
    norm : float (inf)
        Order to use for the norm of the gradient (-np.Inf is min, np.Inf is max).
    maxIter : int (200*len(par))
        Maximum number of iterations
    verbose : bool (False)
        if True print status at convergence
    debug : bool (False)
        return the result of each iteration in `vectors`
    yfit : ndarray (read only)
        the result of the model at the optimal parameters
    ntrans : int (read only)
        number of function calls
    ngrad : int (read only)
        number of gradient calls
    vectors : list of ndarray (read only when debug=True)
        list of intermediate vectors

    Hidden Attributes
    -----------------
    _Chisq : class
        to calculate chisq in the method Chisq.func() and Chisq.dfunc()

    Returns
    -------
    pars : array_like
        the parameters at the minimum of the function (chisq).

    """
## ******************************************************************************
    def __init__( self, xdata, model, method='cg', gradient=True, hessp=None,
                    **kwargs ) :
#                    map=False, keep=None, errdis=None, scale=None, power=2 ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        method : None | 'CG' | 'NELDER-MEAD' | 'POWELL' | 'BFGS' | 'NEWTON-CG' | 'L-BFGS-B' |
                  'TNC' | 'COBYLA' | 'SLSQP' | 'DOGLEG' | 'TRUST-NCG'
            the method name is case invariant.
            None            Automatic selection of the method.
                            'SLSQP' when the problem has constarints
                            'L-BFGS-B' when the problem has limits
                            'BFGS' otherwise
            'CG'            Conjugate Gradient Method of Polak and Ribiere
                            :ref:`(see here) <optimize.minimize-cg>`
            'NELDER-MEAD'   Nelder Mead downhill simplex
                            :ref:`(see here) <optimize.minimize-neldermead>`
            'POWELL'        Powell's conjugate direction method
                            :ref:`(see here) <optimize.minimize-powell>`
            'BFGS'          Quasi-Newton method of Broyden, Fletcher, Goldfarb, and Shannon
                            :ref:`(see here) <optimize.minimize-bfgs>`
            'NEWTON-CG'     Truncated Newton method
                            :ref:`(see here) <optimize.minimize-newtoncg>`
            'L-BFGS-B'      Limited Memory Algorithm for Bound Constrained Optimization
                            :ref:`(see here) <optimize.minimize-lbfgsb>`
            'TNC'           Truncated Newton method with limits
                            :ref:`(see here) <optimize.minimize-tnc>`
            'COBYLA'        Constrained Optimization BY Linear Approximation
                            :ref:`(see here) <optimize.minimize-cobyla>`
            'SLSQP'         Sequential Least Squares
                            :ref:`(see here) <optimize.minimize-slsqp>`
            'DOGLEG'        Dog-leg trust-region algorithm
                            :ref:`(see here) <optimize.minimize-dogleg>`
            'TRUST-NCG'     Newton conjugate gradient trust-region algorithm
                            :ref:`(see here) <optimize.minimize-trustncg>`

        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        hessp : callable hessp(x, p, *args) or None
            Function which computes the Hessian times an arbitrary vector, p.
            The hessian itself is always provided.
        kwargs : dict
            Possibly includes keywords from
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( ScipyFitter, self ).__init__( xdata, model, **kwargs )

        self.userGradient = gradient
        self.userHessp = hessp
        self.method = str.upper( method )
        self.norm = numpy.inf
        self.debug = False
        self.ngrad = 0

        numpy.set_printoptions(formatter={'float': '{: 0.3f}'.format})


    #  *************************************************************************
    def fit( self, data, weights=None, par0=None, keep=None, limits=None,
                maxiter=None, tolerance=None, constraints=(), verbose=0,
                plot=False, callback=None, **options ):
        """
        Return      parameters for the model fitted to the data array.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data
        par0 : array_like
            initial values of the function. Default from Model.
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values of keep are only valid for *this* fit
            See also `ScipyFitter( ..., keep=dict )`
        limits : None or list of 2 floats or list of 2 array_like
            None : no limits applied
            [lo,hi] : low and high limits for all values
            [la,ha] : low array and high array limits for the values
        constraints : list of callables
            constraint functions cf. All are subject to cf(par) > 0.

        maxiter : int
            max number of iterations
        tolerance : float
            stops when ( |hi-lo| / (|hi|+|lo|) ) < tolerance
        verbose : int
            0 : silent
            >0 : print output if iter % verbose == 0
        plot : bool
            Plot the results
        callback : callable
            is called each iteration as
            `val = callback( val )`
            where `val` is the minimizable array
        options : dict
            options to be passed to the method

        Raises
        ------
            ConvergenceError if it stops when the tolerance has not yet been reached.

        """
        fitIndex = self.fitprolog( data, weights=weights, keep=keep )

        inipar = self.model.parameters[fitIndex]

        # Check for limits in the prior
        if limits is not None and self.model.priors :
            try :
                limits = ( self.model.lowLimits[fitIndex], self.model.highLimits[fitIndex] )
            except :
                limits = ( self.model.lowLimits, self.model.highLimits )

        if self.method is None:
            # Select automatically
            if constraints:
                self.method = 'SLSQP'
            elif limits is not None:
                self.method = 'L-BFGS-B'
            else:
                self.method = 'BFGS'

        (func,gradient,hess) = self.makeFuncs( data, weights=weights, index=fitIndex )

        print( self.method )

        # - hessp
        useHessian = self.method in ('NEWTON-CG', 'DOGLEG', 'TRUST-NCG')
        if ( not useHessian and hasattr( self, "userHessp" ) and self.userHessp is not None ) :
            warnings.warn('Method %s does not use Hessian-vector product '
                    'information (hessp).' % self.method, RuntimeWarning)

        # - constraints or limits
        if (self.method in ['NELDER-MEAD', 'POWELL', 'CG', 'BFGS', 'NEWTON-CG', 'DOGLEG',
                 'TRUST-NCG'] and ( limits is not None or numpy.any( constraints ) ) ):
            warnings.warn('Method %s cannot handle constraints nor limits.' % self.method,
                 RuntimeWarning)
        if self.method in ['L-BFGS-B', 'TNC'] and numpy.any( constraints ):
            warnings.warn('Method %s cannot handle constraints.' % self.method, RuntimeWarning)
        if self.method == 'COBYLA' and limits is not None:
            warnings.warn('Method %s cannot handle limits.' % self.method, RuntimeWarning)

        # - callback
        if (self.method in ['COBYLA'] and callback is not None):
            warnings.warn('Method %s does not support callback.' % self.method, RuntimeWarning)
        elif self.debug :
            self.vectors = numpy.asarray( [inipar] )
            print( self.vectors )
            callback = self.collectVectors

        # - return_all
        if (self.method in ['L-BFGS-B', 'TNC', 'COBYLA', 'SLSQP'] and
                options.get('return_all', False)):
            warnings.warn('Method %s does not support the return_all option.' % self.method,
                 RuntimeWarning)

        if options is None:
            options = {}

        if self.method == 'NELDER-MEAD':
            options.setdefault( 'xatol', self.tolerance )
            options.setdefault( 'fatol', self.tolerance )
            options.setdefault( 'maxiter', self.maxIter )
            res = minimize( func, inipar, method="Nelder-Mead",
                    callback=callback, options=options )
        elif self.method == 'POWELL':
            options.setdefault( 'ftol', self.tolerance )
            options.setdefault( 'xtol', self.tolerance )
            options.setdefault( 'maxiter', self.maxIter )
            res = minimize( func, inipar, method="Powell",
                    callback=callback, options=options )
        elif self.method == 'CG':
            options.setdefault( 'gtol', self.tolerance )
            options.setdefault( 'maxiter', self.maxIter )
            res = minimize( func, inipar, jac=gradient, method="CG",
                    callback=callback, options=options )
        elif self.method == 'BFGS':
            options.setdefault( 'gtol', self.tolerance )
            options.setdefault( 'maxiter', self.maxIter )
            res = minimize( func, inipar, jac=gradient, method="BFGS",
                    callback=callback, options=options )
        elif self.method == 'NEWTON-CG':
            options.setdefault( 'xtol', self.tolerance )
            options.setdefault( 'maxiter', self.maxIter )
            res = minimize( func, inipar, jac=gradient, method="Newton-CG",
                    hess=hess, hessp=self.userHessp, callback=callback, options=options )
        elif self.method == 'L-BFGS-B':
            options.setdefault( 'gtol', self.tolerance )
            options.setdefault( 'ftol', self.tolerance )
            res = minimize( func, inipar, jac=gradient, method="L-BFGS-B",
                    bounds=limits, callback=callback, options=options )
        elif self.method == 'TNC':
            options.setdefault( 'gtol', self.tolerance )
            options.setdefault( 'ftol', self.tolerance )
            options.setdefault( 'xtol', self.tolerance )
            res = minimize( func, inipar, jac=gradient, method="TNC",
                    bounds=limits, callback=callback, options=options )
        elif self.method == 'COBYLA':
            options.setdefault( 'tol', self.tolerance )
            res = minimize( func, inipar, method="COBYLA",
                    constraints=constraints, callback=callback, options=options )
        elif self.method == 'SLSQP':
            options.setdefault( 'ftol', self.tolerance )
            res = minimize( func, inipar, jac=gradient, method="SLSQP",
                    bounds=limits, constraints=constraints, callback=callback, options=options )
        elif self.method == 'DOGLEG':
            options.setdefault( 'gtol', self.tolerance )
            res = minimize( func, inipar, jac=gradient, method="dogleg",
                    hess=hess, callback=callback, options=options )
        elif self.method == 'TRUST-NCG':
            options.setdefault( 'gtol', self.tolerance )
            res = minimize( func, inipar, jac=gradient, method="trust-ncg",
                    hess=hess, hessp=self.userHessp, callback=callback, options=options )
        else:
            raise ValueError('Unknown solver %s' % self.method)

        print( res )

        parameters = self.insertParameters( res["x"], index=fitIndex )
        self.model.parameters = parameters

        if self.errdis is None :
            self.chisq = res["fun"]
        else :
            self.logLikelihood = -res["fun"]


#        self.chisq = res["fun"]
        self.ntrans = res["nfev"]
#        print( self.method, self.method != "COBYLA" )
        if self.method != "COBYLA" :
            self.iter = res["nit"]
        # - gradient
        if self.method in ('CG', 'BFGS', 'NEWTON-CG', 'DOGLEG', 'TRUST-NCG') :
            self.ngrad = res["njev"]

        self.fulloutput = res

        warn = res["status"]
        if warn == 1 :
            warnings.warn( "Maximum number of iterations reached" )
        elif warn == 2 :
            warnings.warn( "Gradient and/or function calls were not changing." )

        self.fitpostscript( data, plot=plot )

        return parameters

    def collectVectors( self, par ) :
        self.vectors = numpy.append( self.vectors, [par], 0 )

    def __str__( self ):
        """
        Return the name of the fitter.
        """
        if not hasattr( self, "errdis" ) or self.errdis is None :
            cost = "chisq"
        else :
            cost = self.errdis
        return "ScipyFitter (%s) using %s"%(self.method, cost)

#####################################################################################
###     Specific Fitters                                                        #####
#####################################################################################


class NelderMeadFitter( ScipyFitter ) :
    """
    Nelder Mead downhill simplex.

    Syntactic sugar for `ScipyFitter( ..., method='nelder-mead', ... )`

    """

    def __init__( self, xdata, model, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( NelderMeadFitter, self ).__init__( xdata, model, method='NELDER-MEAD',
                        **kwargs )

class PowellFitter( ScipyFitter ) :
    """
    Powell's conjugate direction method.

    Syntactic sugar for `ScipyFitter( ..., method='powell', ... )`
    """
    def __init__( self, xdata, model, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( PowellFitter, self ).__init__( xdata, model, method='POWELL', **kwargs )

class ConjugateGradientFitter( ScipyFitter ) :
    """
    Conjugate Gradient Method of Polak and Ribiere.

    Syntactic sugar for `ScipyFitter( ..., method='CG', ... )`
    """
    def __init__( self, xdata, model, gradient=True, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( ConjugateGradientFitter, self ).__init__( xdata, model, method='CG',
                **kwargs )

class BfgsFitter( ScipyFitter ) :
    """
    Quasi-Newton method of Broyden, Fletcher, Goldfarb, and Shannon.

    Syntactic sugar for `ScipyFitter( ..., method='BFGS', ... )`
    """
    def __init__( self, xdata, model, gradient=True, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( BfgsFitter, self ).__init__( xdata, model, method='BFGS', gradient=gradient,
                        **kwargs )


class NewtonCgFitter( ScipyFitter ) :
    """
    Truncated Newton method

    Syntactic sugar for `ScipyFitter( ..., method='NEWTON-CG', ... )`
    """
    def __init__( self, xdata, model, gradient=True, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( NewtonCgFitter, self ).__init__( xdata, model, method='NEWTON-CG',
                        gradient=gradient, **kwargs )


class LbfgsbFitter( ScipyFitter ) :
    """
    Limited Memory Algorithm for Bound Constrained Optimization

    Syntactic sugar for `ScipyFitter( ..., method='L-BFGS-B', ... )`
    """
    def __init__( self, xdata, model, gradient=True, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( LbfgsbFitter, self ).__init__( xdata, model, method='L-BFGS-B',
                gradient=gradient, **kwargs )


class TncFitter( ScipyFitter ) :
    """
    Truncated Newton method with limits.

    Syntactic sugar for `ScipyFitter( ..., method='TNC', ... )`
    """
    def __init__( self, xdata, model, gradient=True, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( TncFitter, self ).__init__( xdata, model, method='TNC', gradient=gradient,
                        **kwargs )


class CobylaFitter( ScipyFitter ) :
    """
    Constrained Optimization BY Linear Approximation.

    Syntactic sugar for `ScipyFitter( ..., method='COBYLA', ... )`
    """
    def __init__( self, xdata, model, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( CobylaFitter, self ).__init__( xdata, model, method='COBYLA',
                        **kwargs )


class SlsqpFitter( ScipyFitter ) :
    """
    Sequential Least Squares

    Syntactic sugar for `ScipyFitter( ..., method='SLSQP', ... )`
    """
    def __init__( self, xdata, model, gradient=True, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( SlsqpFitter, self ).__init__( xdata, model, method='SLSQP',
               gradient=gradient, **kwargs )


class DoglegFitter( ScipyFitter ) :
    """
    Dog-leg trust-region algorithm.

    Syntactic sugar for `ScipyFitter( ..., method='DOGLEG', ... )`
    """
    def __init__( self, xdata, model, gradient=True, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( DoglegFitter, self ).__init__( xdata, model, method='DOGLEG',
            gradient=gradient, **kwargs )

class TrustNcgFitter( ScipyFitter ) :
    """
    Newton conjugate gradient trust-region algorithm.

    Syntactic sugar for `ScipyFitter( ..., method='TRUST-NCG', ... )`
    """
    def __init__( self, xdata, model, gradient=True, **kwargs ) :
        """
        Constructor.
        Create a class, providing inputs and model.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            a model function to be fitted (linear or nonlinear)
        gradient : bool or None or callable gradient( par )
            if True use gradient calculated from model. It is the default.
            if False/None dont use gradient (use numeric approximation in stead)
            if callable use the method as gradient
        kwargs : dict
            Possibly includes keywords from
                ScipyFitter:            gradient, hessp
                MaxLikelihoodFitter :   errdis, scale, power
                IterativeFitter :       maxIter, tolerance, verbose
                BaseFitter :            map, keep, fixedScale

        """
        super( TrustNcgFitter, self ).__init__( xdata, model, method='TRUST-NCG',
                        gradient=gradient, **kwargs )





import numpy as numpy
from scipy import special
import math

from .ScaledErrorDistribution import ScaledErrorDistribution
from .HyperParameter import HyperParameter
from .NoiseScale import NoiseScale

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017        Do Kester


class GenGaussErrorDistribution( ScaledErrorDistribution ):
    """
    To calculate a generalized Gaussian likelihood.

    For one residual, x, it holds
    .. math::
        f( x ) = p / ( 2 s \Gamma( 1 / p ) ) exp( - ( |x| / s ) ^ p )

    where s is the scale and p is the power.
    s and p are hyperparameters, which might be estimated from the data.

    The variance of this function is
    .. math::
        \sigma ^ 2 = s ^ 2 \Gamma( 3 / p ) / \Gamma( 1 / p )
    See toSigma()

    The function is mostly used to calculate the likelihood L, or easier
    to use log( L )
    .. math::
        logL = log( N p / ( 2 s \Gamma( 1 / p ) ) ) - \sum( ( |x| / s ) ^ p )

    Using weights this becomes:
    .. math::
        logL = log( \sum( w ) p / ( 2 s \Gamma( 1 / p ) ) ) - \sum( w ( |x| / s ) ^ p )

    Note: the scale s in Generalized Gaussian is NOT the same as the scale in Gaussian or
        in Laplace.

    Author       Do Kester.

    """
    LOG2PI = math.log( 2 * math.pi )
    PARNAMES = ["scale", "power"]

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata, data, weights=None, scale=1.0, power=2.0,
                  limits=None, copy=None ):
        """
        Default Constructor.

        Parameters
        ----------
        xdata : array_like
            input data for the model
        data : array_like
            data to be fitted
        weights : array_like
            weights to be used
        scale : float
            noise scale
        power : float
            power of the distribution
        limits : None or [low,high] or [[low],[high]]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            [low]   low limit on [scale,power] (need to be >0)
            [high]  high limit on [scale,power]
            when limits are set, the scale cq. power are *not* fixed.
        copy : GenGaussErrorDistribution
            distribution to be copied.
        """
        super( GenGaussErrorDistribution, self ).__init__( xdata, data,
                       weights=weights, limits=None, copy=copy )

        plim = None
        if limits is None :
            slim = None
        else :
            lo = limits[0]
            hi = limits[1]
            try :
                slim = [lo[0],hi[0]]
            except :
                slim = [lo,hi]
            try :
                plim = [lo[1],hi[1]]
            except :
                pass

        if copy is None :
            self.hyperpar = [NoiseScale( scale=scale, limits=slim ),
                             HyperParameter( hypar=power, limits=plim )]
        else :
            self.hyperpar = copy.hyperpar

    def copy( self ):
        """ Return copy of this.  """
        return GenGaussErrorDistribution( self.xdata, self.data,
                copy=self )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return True

    def toSigma( self, hypar ) :
        """
        Return sigma, the squareroot of the variance.
        Parameter
        --------
        hypar : array_like (2 floats)
            the [scale,power] of this GenGauss distribution.
        """
        p = hypar[1]
        return hypar[0] * math.sqrt( special.gamma( 3.0 / p ) / special.gamma( 1.0 / p ) )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood( self, model, allpars ) :
        """
        Return the log( likelihood ) for a Gaussian distribution.

        Parameters
        ----------
        model : Model
            model to calculate mock data
        allpars : array_like
            parameters of the problem

        """
        self.ncalls += 1
        np = model.npchain
        scale = allpars[np]
        power = allpars[np+1]

        res = self.getResiduals( model, allpars[:np] )
        chisq = self.getChisq( res, scale, power )
        norm = math.log( power / ( 2 * scale ) ) - special.gammaln( 1.0 / power )
#        print( "GG  ", chisq, norm, self.sumweight )
        return self.sumweight * norm - chisq

    def getChisq( self, residual, scale, power ):
        """
        Return chisq.

        Sum over the (weighted) squared residuals

        Parameters
        ----------
        residual : array_like
            the residuals
        scale : float
            noise scale
        power : float
            power of distribution
        """
        ares = numpy.power( numpy.abs( residual / scale ), power )
        if self.weights is not None :
            ares = ares * self.weights
        return numpy.sum( ares )

    def getScale( self, model ) :
        """
        Return the noise scale calculated from the residuals.

        Parameters
        ----------
        model : Model
            the model involved.
        """
        power = self.hyperpar[1].hypar
        chi = self.getChisq( self.getResiduals( model ), 1.0, power )
        return math.pow( chi / self.sumweight, 1.0 / power )


    def partialLogL( self, model, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Parameters
        ----------
        model : Model
            model to calculate mock data
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of the parameters to be fitted

        """
        self.ncalls += 1
        np = model.npchain
        scale = allpars[np]
        power = allpars[np+1]
        res = self.getResiduals( model, allpars[:np] )

        ars = numpy.abs( res / scale )
        rsp = numpy.power( ars, power )
        if self.weights is not None :
            rsp = rsp * self.weights

        dLdm = power * rsp / res
        dM = model.partial( self.xdata, allpars[:np] )

        dL = numpy.zeros( len( fitIndex ), dtype=float )
        i = 0
        for  k in fitIndex :
            if k < np :
                dL[i] = numpy.sum( dLdm * dM[:,k] )
            elif k == np :
                dL[i] = - self.sumweight / scale + power * numpy.sum( rsp ) / scale
            else :
                # special.psi( x ) is the same as special.polygamma( 1, x )
                dL[i] = self.sumweight * ( power + special.psi( 1.0 / power ) )
                dL[i] /= ( power * power )
                dL[i] -= ( numpy.sum( rsp * numpy.log( ars ) ) )
            i += 1

        return dL

    def __str__( self ) :
        return "Generalized Gauss error distribution"




# run with : python3 -m unittest TestModels.TestModels.testGaussModel

import unittest
import numpy as numpy
from astropy import units
import matplotlib.pyplot as plt
import warnings

from StdTests import stdModeltest

from BayesicFitting import *


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestModels( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def testToString( self ) :
        model = PadeModel( 1, 2 )
        model += PolynomialModel( 0 )
        name = model.__str__()
        print( name )
        self.assertTrue( name[-1] == '4' )

        model = PadeModel( 1, 2 )
        model += PolynomialModel( 1 )
        print( model )

    def plotArctanModel( self ) :
        self.testArctanModel( plot=True )

    def plotConstantModel( self ) :
        self.testConstantModel( plot=True )

    def plotEtalonModel( self ) :
        self.testEtalonModel( plot=True )

    def plotExpModel( self ) :
        self.testExpModel( plot=True )

    def plotVoigtModel( self ) :
        self.testVoigtModel( )
        xx = numpy.linspace( -1, +1, 1001 )
        par = numpy.asarray( [1.2,-0.2,0.3,0.3], dtype=float )
        plt.plot( xx, VoigtModel().result( xx, par ), '-', linewidth=2 )
        plt.plot( xx, GaussModel().result( xx, par[[0,1,2]] ), 'r-' )
        pg = par.copy()
        pg[3] = 0
        plt.plot( xx, VoigtModel().result( xx, pg ), 'r--' )
        plt.plot( xx, LorentzModel().result( xx, par[[0,1,3]] ), 'g-' )
        pg = par.copy()
        pg[2] = 0
        plt.plot( xx, VoigtModel().result( xx, pg ), 'g--' )
        plt.show()

    def plotLorentzModel( self ) :
        self.testLorentzModel( plot=True )

    def plotFreeShapeModel( self ) :
        self.XtestFreeShapeModel( plot=True )

    def plotGaussModel( self ) :
        self.testGaussModel( plot=True )

    def plotGaussPlusBackgroundModel( self ) :
        self.testGaussPlusBackgroundModel( plot=True )

    def plotHarmonicModel( self ) :
        self.testHarmonicModel( plot=True )

    def plotPolynomialModel( self ) :
        self.testPolynomialModel( plot=True )

    def plotChebyshevPolynomialModel( self ) :
        self.testChebyshevPolynomialModel( plot=True )

    def plotPadeModel( self ) :
        self.testPadeModel( plot=True )

    def plotPowerLawModel( self ) :
        self.testPowerLawModel( plot=True )

    def plotPowerModel( self ) :
        self.testPowerModel( plot=True )

    def plotSincModel( self ) :
        self.testSincModel( plot=True )

    def plotSineModel( self ) :
        self.testSineModel( plot=True )

    def plotSineAmpModel( self ) :
        self.testSineAmpModel( plot=True )

    def plotSineSplineModel( self ) :
        self.testSineSplineModel( plot=True )

    def plotSineSplineDriftModel( self ) :
        self.testSineSplineDriftModel( plot=True )

    def plotSineDriftModel( self ) :
        self.testSineDriftModel( plot=True )

    def plotBSplinesModel( self ) :
        self.testBSplinesModel( plot=True )

    def plotSplinesModel( self ) :
        self.testSplinesModel( plot=True )


    def testArctanModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ARCTAN***********************" )
        m = ArctanModel( )
        p = numpy.asarray( [1.2,-0.1,30], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testArctanFixedModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ARCTAN FIXED******************" )
        m = ArctanModel( fixed={1:-1.2} )
        p = numpy.asarray( [1.2,30], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testConstantModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******NULL***********************" )
        m = ConstantModel( )
        self.assertTrue( m.npchain == 0 )
        p = []
        stdModeltest( m, p )

        print( "******CONSTANT*******************" )
        m = ConstantModel( values = 5.5 )
        stdModeltest( m, p )

        print( "******COMPOUND*******************" )
        m = ConstantModel( values = 1.0 )
        m.addModel( ExpModel() )
        self.assertTrue( m.npbase == 0 )
        self.assertTrue( m.npchain == 2 )
        p = [1.0, -3.0]
        stdModeltest( m, p, plot=plot )

    def testEtalonModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ETALON***********************" )
        m = EtalonModel( )
        p = numpy.asarray( [1.2, 0.6, 2.0, 0.2], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testExpModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******EXP**************************" )
        m = ExpModel( )
        p = numpy.asarray( [1.2,-0.1], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testLorentzModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******LORENTZ***********************" )
        m = LorentzModel( )
        p = numpy.asarray( [1.2,-0.2,0.3], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testVoigtModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******VOIGT***********************" )
        m = VoigtModel( )
        p = numpy.asarray( [1.2,0.2,0.3,0.4], dtype=float )
        print( p )
        stdModeltest( m, p, plot=plot, warn=["nopart"] )

    def XtestFreeShapeModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******FREESHAPE********************" )
        m = FreeShapeModel( 5, pixperbin=1, xlo=-1, xhi=1.1 )
        self.assertTrue( m.npbase == 5 )
        p = numpy.asarray( [1.2,0.2,1.0,0.5,0.3], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testGaussModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******GAUSS*************************" )
        m = GaussModel( )
        p = numpy.asarray( [1.2,-0.2,0.3], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testGaussPlusBackgroundModel( self, plot=False ):
        print( "******GAUSS + BG**********************" )
        gm = GaussModel( )
        print( gm )
        print( gm.parameters )
        pm = PolynomialModel( 1 )
        print( pm )
        print( pm.parameters )
        gm.addModel( pm )
        par = numpy.asarray( [3,0.2,0.2,0.1,0.1], dtype=float )

        stdModeltest( gm, par, plot=plot )

    def testHarmonicModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******HARMONIC**********************" )
        m = HarmonicModel( 1 )
        self.assertTrue( m.order == 1 )
        self.assertTrue( m.npbase == 2 )
        self.assertTrue( m.period == 1.0 )

        m = HarmonicModel( order=3, period=0.4 )
        self.assertTrue( m.order == 3 )
        self.assertTrue( m.npbase == 6 )
        self.assertTrue( m.period == 0.4 )

        p = numpy.asarray( [1.2,-0.2,0.3,1.2,-0.2,0.3], dtype=float )

        stdModeltest( m, p, plot=plot )

    def testPolynomialModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POLYNOMIAL**********************" )
        m = PolynomialModel( 3 )
        self.assertTrue( m.getNumberOfParameters( ) == 4 )
        self.assertTrue( m.npbase == 4 )
        p = numpy.asarray( [1,-2,3,-2], dtype=float )
        stdModeltest( m, p, plot=plot )

    def testFixedPolynomialModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POLYNOMIAL FIXED****************" )
        m = PolynomialModel( 5, fixed={1:0.4, 3:-1.2} )
        self.assertTrue( m.getNumberOfParameters( ) == 4 )
        self.assertTrue( m.npbase == 4 )
        self.assertTrue( m.npmax == 6 )
        p = numpy.asarray( [1,-2,3,-2], dtype=float )
        stdModeltest( m, p, plot=plot )

    def testChebyshevPolynomialModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******CHEBYSHEV POLYNOMIAL*************" )
        m = ChebyshevPolynomialModel( 4 )
        self.assertTrue( m.getNumberOfParameters( ) == 5 )
        self.assertTrue( m.npbase == 5 )
        p = numpy.asarray( [1,-2,3,-2,0.3], dtype=float )
        stdModeltest( m, p, plot=plot )

    def testPadeModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******PADE**********************" )
        m = PadeModel( 3, 2 )
        self.assertTrue( m.getNumberOfParameters( ) == 6 )
        self.assertTrue( m.npbase == 6 )
        p =  numpy.asarray( [1,-2,3,-2,0.3,1], dtype=float )
        stdModeltest( m, p, plot=plot )

    def testPowerLawModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POWERLAW*******************" )
        m = PowerLawModel( )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [2.3, -1.1, 0.5]
        stdModeltest( m, p, plot=plot )

    def testPowerModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POWER**********************" )
        m = PowerModel( 3 )
        self.assertTrue( m.getNumberOfParameters( ) == 1 )
        self.assertTrue( m.npbase == 1 )
        p = [2.3]
        stdModeltest( m, p, plot=plot )

    def testSincModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINC***********************" )
        m = SincModel( )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [2.3, -1.1, 0.5]
        stdModeltest( m, p, plot=plot )

    def testSineModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINE***********************" )
        m = SineModel( )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [1.3, -1.1, 0.5]
        stdModeltest( m, p, plot=plot )

    def testSineAmpModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINEAMP********************" )
        m = SineAmpModel( 1.3 )
        self.assertTrue( m.npchain == 2 )
        self.assertTrue( m.npbase == 2 )
        p = [-1.1, 0.5]
        stdModeltest( m, p, plot=plot )

    def testSineSplineModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINESPLINE******************" )
        knots = numpy.arange( 3, dtype=float ) - 1
        m = SineSplineModel( 1.3, knots )
        self.assertTrue( m.npchain == 10 )
        self.assertTrue( m.npbase == 10 )
        p = [0.0, 0.5, -0.1, 0.5, -0.1, 0.0, +1.1, 0.5, -1.1, -0.5]
        amp = m.getAmplitudes( x, p )
        print( amp[0], amp[1] )
        stdModeltest( m, p, plot=plot )

    def testSineSplineDriftModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINESPLINEDRIFT*************" )
        knots = numpy.arange( 3, dtype=float ) - 1
        m = SineSplineDriftModel( knots, degree=1 )
        print( m.npchain )
        self.assertTrue( m.npchain == 12 )
        self.assertTrue( m.npbase == 12 )
        p = [0.0, 0.1, 0.0, 0.5, -0.1, 0.5, -0.1, 0.0, +1.1, 0.5, -1.1, -0.5]
        amp = m.getAmplitudes( x, p )
        print( amp[0], amp[1] )
        stdModeltest( m, p, plot=plot )

    def testSineDriftModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINEDRIFT******************" )
        m = SineDriftModel( )
        self.assertTrue( m.npchain == 4 )
        self.assertTrue( m.npbase == 4 )
        p = [1.1, 0.5, 0.8, 0.4 ]
        stdModeltest( m, p, plot=plot )

    def testBSplinesModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******B-SPLINES*************************" )
        knots = numpy.arange( 11, dtype=float )
        m0 = BSplinesModel( knots=knots )
        self.assertTrue( m0.order == 3 )
        self.assertTrue( m0.getNumberOfParameters( ) == 13 )
        self.assertTrue( m0.npbase == 13 )
        m1 = BSplinesModel( nrknots=5, min=-1, max=1 )
        self.assertTrue( m1.order == 3 )
        self.assertTrue( m1.getNumberOfParameters( ) == 7 )
        self.assertTrue( m1.npbase == 7 )
        for k in range( len( m1.knots ) ) :
            self.assertTrue( m1.knots[k] == -1 + k / 2 )
        xin = numpy.linspace( 0, 10, 101, dtype = float )
        m2 = BSplinesModel( xrange=xin, order=2, nrknots=11 )
        self.assertTrue( m2.order == 2 )
        self.assertTrue( m2.getNumberOfParameters( ) == 12 )
        self.assertTrue( m2.npbase == 12 )
        for k in range( len( m2.knots ) ) : self.assertTrue( m2.knots[k] == k )

        self.assertRaises( ValueError, BSplinesModel )

        m1 = BSplinesModel( nrknots=5, order=3, min=-1, max=1 )
        p = numpy.asarray( [2,1,2,1,2,1,2], dtype=float )
        p = p[:m1.npchain]
        print( m1.npchain, p )
        stdModeltest( m1, p, plot=plot )

    def testSplinesModel( self, plot=False ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SPLINES*************************" )
        knots = numpy.arange( 11, dtype=float )
        m0 = SplinesModel( knots=knots )
        self.assertTrue( m0.order == 3 )
        self.assertTrue( m0.getNumberOfParameters( ) == 13 )
        self.assertTrue( m0.npbase == 13 )
        m1 = SplinesModel( nrknots=5, min=-1, max=1 )
        self.assertTrue( m1.order == 3 )
        self.assertTrue( m1.getNumberOfParameters( ) == 7 )
        self.assertTrue( m1.npbase == 7 )
        for k in range( len( m1.knots ) ) :
            self.assertTrue( m1.knots[k] == -1 + k / 2 )
        xin = numpy.linspace( 0, 10, 101, dtype = float )
        m2 = SplinesModel( xrange=xin, order=2, nrknots=11 )
        self.assertTrue( m2.order == 2 )
        self.assertTrue( m2.getNumberOfParameters( ) == 12 )
        self.assertTrue( m2.npbase == 12 )
        for k in range( len( m2.knots ) ) : self.assertTrue( m2.knots[k] == k )

        self.assertRaises( ValueError, SplinesModel )

        p = numpy.asarray( [0,1,0,1,0,1,0], dtype=float )
        stdModeltest( m1, p, plot=plot )



    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestModels.__class__ )

if __name__ == '__main__':
    unittest.main( )


